from __future__ import annotations

import threading
from collections.abc import Generator
from dataclasses import replace
from http.server import BaseHTTPRequestHandler, HTTPServer
from pathlib import Path
from typing import Any

import pytest

from screenshot.models import CaptureOptions, ScreenshotOptions
from screenshot.playwright_runner import capture_screenshots_async


class _SmokeHandler(BaseHTTPRequestHandler):
    def do_GET(self) -> None:  # noqa: N802 - signature defined by BaseHTTPRequestHandler
        body = (
            "<html><head><title>Smoke</title></head>"
            "<body><h1>Smoke Test</h1><p id='content'>ok</p></body></html>"
        )
        self.send_response(200)
        self.send_header("Content-Type", "text/html; charset=utf-8")
        self.end_headers()
        self.wfile.write(body.encode("utf-8"))

    def log_message(self, format: str, *args: object) -> None:  # noqa: A003
        return  # keep pytest output clean


@pytest.fixture(scope="module")
def smoke_site() -> Generator[str, None, None]:
    try:
        server = HTTPServer(("127.0.0.1", 0), _SmokeHandler)
    except PermissionError:
        pytest.skip(
            "Cannot bind HTTP server inside this environment; smoke test needs network access"
        )
    thread = threading.Thread(target=server.serve_forever, daemon=True)
    thread.start()
    base = f"http://127.0.0.1:{server.server_address[1]}"
    try:
        yield base
    finally:
        server.shutdown()
        thread.join()


def _smoke_options(**capture_overrides: Any) -> ScreenshotOptions:
    capture = CaptureOptions(
        enabled=True,
        max_pages=1,
        depth=0,
        viewports=("desktop",),
        post_nav_wait_s=0.1,
        timeout_s=20.0,
        max_capture_attempts=1,
    )
    if capture_overrides:
        capture = replace(capture, **capture_overrides)
    return ScreenshotOptions(capture=capture)


@pytest.mark.asyncio
async def test_capture_local_site_smoke(tmp_path: Path, smoke_site: str) -> None:
    pytest.importorskip("playwright.async_api")

    result = await capture_screenshots_async(
        "smoke",
        f"{smoke_site}/",
        store_dir=tmp_path,
        partition_date=None,
        options=_smoke_options(),
    )

    if result.errors and any("playwright" in err.message.lower() for err in result.errors):
        pytest.skip(
            "Playwright capture failed; run `playwright install chromium` "
            "(and ensure the browsers are available) before rerunning the smoke test."
        )

    assert result.succeeded
    assert result.metadata_path and result.metadata_path.exists()
    entries = result.entries
    assert entries, "Expected metadata entries for smoke job"


@pytest.mark.asyncio
async def test_capture_failure_records_errors(tmp_path: Path) -> None:
    pytest.importorskip("playwright.async_api")

    bad_url = "http://127.0.0.1:9/does-not-exist"
    result = await capture_screenshots_async(
        "smoke-failure",
        bad_url,
        store_dir=tmp_path,
        partition_date=None,
        options=_smoke_options(timeout_s=5.0),
    )

    assert not result.succeeded
    assert result.errors or any(
        isinstance(entry, dict) and entry.get("error") for entry in result.entries
    )



"""Shared helpers for screenshot CLI commands."""

from __future__ import annotations

from typing import NotRequired, TypedDict

from ._shared.cli.options import build_options
from ._shared.cli.parser import load_cli_inputs
from ._shared.cli.schema import ScreenshotCliArgs
from ._shared.cli.utils import derive_job_id
from ._shared.cli.validator import validate_backend_choice
from ._shared.options_normalizer import ensure_sequence_lists
from .models import ScreenshotOptions


class ScreenshotSpecMetadata(TypedDict, total=False):
    """Normalized metadata attached to each CLI job spec."""

    source: NotRequired[str]
    backend: NotRequired[str]
    schema_version: NotRequired[str]
    campaign: NotRequired[str]


class ScreenshotJobSpec(TypedDict):
    """Explicit schema for the job spec returned by `collect_job_specs`."""

    job_id: str
    url: str
    partition_date: str | None
    html_snapshot_path: str | None
    metadata: ScreenshotSpecMetadata
    options: ScreenshotOptions
    backend: str


def collect_job_specs(args: ScreenshotCliArgs) -> list[ScreenshotJobSpec]:
    """Return normalized screenshot job specs from CLI arguments."""

    backend = (args.backend or "playwright").strip().lower()
    records, css_snippets, js_snippets = load_cli_inputs(args)

    specs: list[ScreenshotJobSpec] = []
    for record in records:
        options = build_options(
            args=args,
            css_snippets=css_snippets,
            js_snippets=js_snippets,
            overrides=record.options_override,
        )
        validate_backend_choice(backend, options)

        metadata: ScreenshotSpecMetadata = dict(record.metadata)  # type: ignore[assignment]
        metadata.setdefault("backend", backend)

        specs.append(
            {
                "job_id": record.job_id,
                "url": record.url,
                "partition_date": record.partition_date,
                "html_snapshot_path": record.html_snapshot_path,
                "metadata": metadata,
                "options": options,
                "backend": backend,
            }
        )

    return specs


def spec_to_api_payload(spec: ScreenshotJobSpec) -> dict[str, object]:
    """Convert a job spec into the payload expected by the screenshot service."""

    payload: dict[str, object] = {
        "job_id": spec["job_id"],
        "url": spec["url"],
        "partition_date": spec["partition_date"],
        "metadata": dict(spec["metadata"]),
        "options": serialize_options(spec["options"]),
    }
    backend_value = spec.get("backend")
    if backend_value:
        payload["backend"] = backend_value
    snapshot = spec.get("html_snapshot_path")
    if snapshot:
        payload["html_snapshot_path"] = snapshot
    return payload


def serialize_options(options: ScreenshotOptions) -> dict[str, object]:
    """Serialize ScreenshotOptions to simple types for JSON payloads."""

    normalized = ensure_sequence_lists(options)
    payload = normalized.to_dict()
    return payload


__all__ = [
    "collect_job_specs",
    "spec_to_api_payload",
    "serialize_options",
    "derive_job_id",
    "ScreenshotCliArgs",
]



from __future__ import annotations

from pathlib import Path

import pytest

from screenshot._models_options import ScreenshotOptions
from screenshot._shared.options_normalizer import (
    _normalize_viewports,
    ensure_capture_enabled,
    ensure_sequence_lists,
    normalize_options_payload,
)


def test_normalize_options_payload_flat_schema() -> None:
    raw = {
        "enabled": True,
        "max_pages": 3,
        "depth": 1,
        "viewports": ("desktop", "mobile"),
        "delay": 0.5,
        "timeout": 10.0,
        "allow_autoplay": False,
        "block_media": True,
        "playwright_executable_path": "/opt/chrome",
        "extra": {"feature": "on"},
        "schema_version": "ignored-here",
    }

    normalized = normalize_options_payload(raw)

    assert normalized["capture"]["enabled"] is True
    assert normalized["capture"]["post_nav_wait_s"] == 0.5
    assert normalized["capture"]["timeout_s"] == 10.0
    assert normalized["capture"]["viewports"] == ["desktop", "mobile"]
    assert normalized["browser"]["block_media"] is True
    assert normalized["browser"]["allow_autoplay"] is False
    assert normalized["runner"]["playwright_executable_path"] == "/opt/chrome"
    assert normalized["runner"]["extra"] == {"feature": "on"}


def test_normalize_options_payload_nested_copy_safe() -> None:
    raw = {"capture": {"viewports": ["desktop"]}}
    normalized = normalize_options_payload(raw)

    assert normalized["capture"]["viewports"] == ["desktop"]
    assert normalized is not raw
    assert normalized["capture"] is not raw["capture"]
    normalized["capture"]["viewports"].append("tablet")
    assert raw["capture"]["viewports"] == ["desktop"]


def test_normalize_viewports_favors_non_empty_entries() -> None:
    assert _normalize_viewports(None) == []
    assert _normalize_viewports(" mobile ") == ["mobile"]
    assert _normalize_viewports({"name": "custom"}) == [{"name": "custom"}]
    assert _normalize_viewports({"key": "value"}) == [{"key": "value"}]
    assert _normalize_viewports({"name": None}) == [{"name": None}]
    assert _normalize_viewports(["desktop", None, "mobile"]) == ["desktop", "mobile"]


def test_ensure_capture_enabled_and_sequence_lists() -> None:
    options = ScreenshotOptions.from_dict({"capture": {"enabled": False}})
    enabled = ensure_capture_enabled(options)
    assert enabled.capture.enabled is True
    assert options.capture.enabled is False

    expanded = ScreenshotOptions.from_dict(
        {
            "capture": {"enabled": True, "viewports": ["desktop", "mobile"]},
            "runner": {
                "extra_styles": ["body { color: red; }"],
                "extra_init_scripts": ["console.log('init')"],
                "extra": {"trace": True},
                "playwright_executable_path": Path("/usr/bin/chrome"),
            },
        }
    )
    flattened = ensure_sequence_lists(expanded)
    assert flattened.capture.viewports == ("desktop", "mobile")
    assert list(flattened.runner.extra_styles) == ["body { color: red; }"]
    assert list(flattened.runner.extra_init_scripts) == ["console.log('init')"]
    assert flattened.runner.extra["trace"] is True
    assert flattened.runner.playwright_executable_path == str(Path("/usr/bin/chrome"))


def test_schema_version_passthrough() -> None:
    with pytest.raises(ValueError):
        ScreenshotOptions.from_dict({"schema_version": "unexpected"})

    valid = ScreenshotOptions.from_dict({"capture": {"enabled": True}})
    assert valid.to_dict()["schema_version"] == "screenshot_options/v2"



from __future__ import annotations

from test_utils import create_capture_result

from screenshot import ScreenshotCaptureResult, ScreenshotResourceOutcome
from screenshot._internal import from_capture_result, per_viewport_resources
from screenshot._shared.errors import ScreenshotError


def test_from_capture_result_success_metadata_and_config_normalization() -> None:
    capture = create_capture_result(requested=2, captured=2, failed=0)

    config_a = {
        "viewports": ["desktop", "mobile"],
        "depth": 1,
        "timeout_s": 45.0,
        "scroll": True,
        "full_page": False,
    }
    config_b = {
        "viewports": ["mobile", "desktop"],  # different order but equivalent
        "depth": 1,
        "timeout_s": 45.0,
        "scroll": True,
        "full_page": False,
        "ignored": "value",
    }

    result_a = from_capture_result(
        capture,
        resource_key="screenshots",
        config=config_a,
    )
    result_b = from_capture_result(
        capture,
        resource_key="screenshots",
        config=config_b,
    )

    assert result_a.outcome is ScreenshotResourceOutcome.SUCCESS
    assert result_a.errors == []
    assert result_a.metadata["requested"] == 2
    assert result_a.metadata["captured"] == 2
    assert result_a.metadata["job_id"] == "job-123"
    assert result_a.metadata["metadata_path"].endswith("meta.json")
    assert result_a.config_key == result_b.config_key


def test_from_capture_result_failure_and_error_categorization() -> None:
    capture = create_capture_result(
        requested=1,
        captured=0,
        failed=1,
        errors=[
            ScreenshotError(error_type="TimeoutError", message="took too long", retryable=True),
            ScreenshotError(error_type="NavigationCrash", message="nav failed"),
        ],
    )

    result = from_capture_result(
        capture,
        resource_key="screenshots.desktop",
        config={"viewports": ["desktop"]},
    )

    assert result.outcome is ScreenshotResourceOutcome.FAILED
    assert [err.category for err in result.errors] == [
        "screenshot.timeout",
        "screenshot.navigation",
    ]
    assert result.errors[0].retryable is True


def test_from_capture_result_partial_and_custom_metadata() -> None:
    capture = create_capture_result(requested=3, captured=2, failed=1)

    result = from_capture_result(
        capture,
        resource_key="screenshots.bundle",
        config={"viewports": ["desktop", "tablet"]},
        metadata={"extra": "value"},
    )

    assert result.outcome is ScreenshotResourceOutcome.PARTIAL
    assert result.metadata["extra"] == "value"


def test_from_capture_result_skipped_when_no_requests() -> None:
    capture = create_capture_result(requested=0, captured=0, failed=0)

    result = from_capture_result(
        capture,
        resource_key="screenshots",
        config={"viewports": []},
    )

    assert result.outcome is ScreenshotResourceOutcome.SKIPPED


def test_per_viewport_resources_breaks_out_entries() -> None:
    capture = ScreenshotCaptureResult(
        requested=2,
        captured=1,
        failed=1,
        metadata_path=None,
        entries=[
            {"viewport": "desktop", "status": "success"},
            {"viewport": "mobile", "status": "failed", "error": "timeout"},
        ],
        errors=[
            ScreenshotError(
                error_type="TimeoutError",
                message="timed out",
                retryable=True,
            )
        ],
    )

    result = from_capture_result(
        capture,
        resource_key="screenshots",
        config={"viewports": ["desktop", "mobile"]},
    )

    viewports = per_viewport_resources(result)
    names = {v.viewport: v for v in viewports}
    assert set(names.keys()) == {"desktop", "mobile"}
    assert names["desktop"].outcome is ScreenshotResourceOutcome.SUCCESS
    assert names["mobile"].outcome is ScreenshotResourceOutcome.FAILED
    assert names["mobile"].metadata["failed"] == 1



from __future__ import annotations

import math

import pytest

from screenshot import defaults


def test_defaults_respect_env_overrides(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("SCREENSHOT_SCROLL_STEP_DELAY_MS", "123")
    monkeypatch.setenv("SCREENSHOT_MAX_SCROLL_STEPS", "5")
    monkeypatch.setenv("PLAYWRIGHT_CAPTURE_MAX_ATTEMPTS", "7")
    monkeypatch.setenv("SCREENSHOT_RETRY_BACKOFF_S", "0.75")
    monkeypatch.setenv("SCREENSHOT_RETRY_MAX_BACKOFF_S", "9.5")
    monkeypatch.setenv("SCREENSHOT_ENABLE_TIMING", "true")

    loaded = defaults.ScreenshotDefaults.load()

    assert loaded.scroll_step_delay_ms == 123
    assert loaded.max_scroll_steps == 5
    assert loaded.playwright_max_capture_attempts == 7
    assert loaded.retry_backoff_s == 0.75
    assert loaded.retry_max_backoff_s == 9.5
    assert loaded.enable_timing is True


def test_defaults_fall_back_on_invalid_env(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("SCREENSHOT_SCROLL_STEP_DELAY_MS", "not-an-int")
    monkeypatch.setenv("SCREENSHOT_RETRY_BACKOFF_S", "NaN")

    loaded = defaults.ScreenshotDefaults.load()

    assert loaded.scroll_step_delay_ms == 250
    assert math.isnan(loaded.retry_backoff_s)



import json
from pathlib import Path
from typing import cast

import pytest

from screenshot import ScreenshotOptions
from screenshot._models_options import (
    OPTIONS_SCHEMA_VERSION,
    BrowserCompatOptions,
    CaptureOptions,
    RunnerOptions,
)
from screenshot.cli_utils import (
    ScreenshotCliArgs,
    ScreenshotJobSpec,
    collect_job_specs,
    derive_job_id,
    serialize_options,
    spec_to_api_payload,
)


def test_collect_job_specs_from_args(tmp_path: Path) -> None:
    cli_args = ScreenshotCliArgs(
        input=None,
        urls=("https://example.com",),
        site_ids=(),
        partition_date="2025-10-21",
        max_pages=3,
        depth=1,
        viewports=("desktop",),
        post_nav_wait_s=0.5,
        timeout_s=45.0,
        max_retries=2,
        job_budget_s=180.0,
        max_viewport_concurrency=None,
    )

    specs = collect_job_specs(cli_args)
    assert len(specs) == 1
    spec: ScreenshotJobSpec = specs[0]
    assert spec["partition_date"] == "2025-10-21"
    assert spec["options"].capture.max_pages == 3
    assert spec["options"].capture.max_total_duration_s == 180.0
    assert spec["options"].browser.allow_autoplay is True
    assert spec["options"].browser.hide_overlays is True
    assert spec["options"].browser.reduced_motion is True
    assert spec["options"].browser.mute_media is True
    assert spec["options"].browser.disable_animations is True
    assert spec["options"].browser.compatibility_level == "high"
    assert spec["options"].capture.max_viewport_concurrency == 1
    assert spec["options"].browser.user_agent is None
    assert spec["metadata"]["source"] == "cli"
    assert spec["metadata"]["backend"] == "playwright"
    assert spec["backend"] == "playwright"
    assert spec["options"].runner.playwright_executable_path is None

    payload: dict[str, object] = spec_to_api_payload(spec)
    assert payload["job_id"]
    options_payload = cast(dict[str, object], payload["options"])
    capture_payload = cast(dict[str, object], options_payload["capture"])
    browser_payload = cast(dict[str, object], options_payload["browser"])
    assert capture_payload["max_pages"] == 3
    assert browser_payload["allow_autoplay"] is True
    assert browser_payload["mute_media"] is True
    assert browser_payload.get("user_agent") is None
    assert capture_payload["max_viewport_concurrency"] == 1
    assert payload["backend"] == "playwright"


def test_collect_job_specs_uses_scroll_tuning_args(tmp_path: Path) -> None:
    cli_args = ScreenshotCliArgs(
        input=None,
        urls=("https://example.com",),
        site_ids=(),
        partition_date="2025-10-21",
        max_pages=1,
        depth=0,
        viewports=("desktop",),
        post_nav_wait_s=0.5,
        timeout_s=45.0,
        max_retries=1,
        job_budget_s=None,
        scroll_step_delay_ms=180,
        max_scroll_steps=7,
        max_viewport_concurrency=None,
    )

    specs = collect_job_specs(cli_args)
    options = specs[0]["options"]

    assert options.capture.scroll_step_delay_ms == 180
    assert options.capture.max_scroll_steps == 7


def test_collect_job_specs_from_json(tmp_path: Path) -> None:
    input_path = tmp_path / "jobs.jsonl"
    input_path.write_text(
        json.dumps(
            {
                "job_id": "marketing-home",
                "url": "https://example.com",
                "partition_date": "2025-10-22",
                "options": {
                    "capture": {
                        "enabled": True,
                        "max_pages": 4,
                        "viewports": ["desktop", "mobile"],
                    },
                    "browser": {},
                    "runner": {},
                },
                "metadata": {"campaign": "launch"},
                "user_agent": "FromFile/1.0",
            }
        )
        + "\n",
        encoding="utf-8",
    )

    cli_args = ScreenshotCliArgs(
        input=input_path,
        urls=(),
        site_ids=(),
        partition_date=None,
        max_pages=5,
        depth=1,
        viewports=("desktop",),
        post_nav_wait_s=0.0,
        timeout_s=60.0,
        max_retries=None,
        job_budget_s=None,
        scroll=False,
        max_viewport_concurrency=None,
    )

    specs = collect_job_specs(cli_args)
    assert len(specs) == 1
    spec: ScreenshotJobSpec = specs[0]
    assert spec["job_id"] == "marketing-home"
    assert spec["options"].capture.max_pages == 4
    assert "mobile" in spec["options"].capture.viewports
    assert spec["options"].browser.allow_autoplay is True
    assert (
        spec["options"].browser.user_agent is None
    )  # legacy top-level user_agent field is ignored
    assert spec["metadata"]["campaign"] == "launch"
    assert spec["metadata"]["backend"] == "playwright"
    assert spec["backend"] == "playwright"


def test_collect_job_specs_from_nested_json(tmp_path: Path) -> None:
    input_path = tmp_path / "jobs_v2.jsonl"
    input_path.write_text(
        json.dumps(
            {
                "job_id": "blog-detail",
                "url": "https://example.com/blog/intro",
                "partition_date": "2025-11-01",
                "options": {
                    "schema_version": OPTIONS_SCHEMA_VERSION,
                    "capture": {
                        "enabled": False,
                        "max_pages": 2,
                        "depth": 2,
                        "viewports": ["mobile"],
                        "post_nav_wait_s": 1.25,
                        "max_total_duration_s": 40.0,
                        "scroll": False,
                        "pre_capture_wait_s": 0.5,
                    },
                    "browser": {
                        "allow_autoplay": False,
                        "hide_overlays": False,
                        "reduced_motion": True,
                        "compatibility_level": "medium",
                        "user_agent": "NestedUA/1.0",
                    },
                    "runner": {
                        "extra_styles": ["body { color: green; }"],
                        "extra_init_scripts": ["console.log('nested');"],
                        "extra": {"locale": "fr-FR"},
                        "playwright_executable_path": "/usr/bin/google-chrome-stable",
                    },
                },
            }
        )
        + "\n",
        encoding="utf-8",
    )

    cli_args = ScreenshotCliArgs(
        input=input_path,
        urls=(),
        site_ids=(),
        partition_date=None,
        max_pages=5,
        depth=1,
        viewports=("desktop",),
        post_nav_wait_s=0.0,
        timeout_s=45.0,
        max_retries=None,
        job_budget_s=None,
        scroll=True,
        allow_autoplay=True,
        hide_overlays=True,
        reduced_motion=False,
    )

    specs = collect_job_specs(cli_args)
    assert len(specs) == 1
    spec: ScreenshotJobSpec = specs[0]
    options = spec["options"]
    assert spec["backend"] == "playwright"
    assert spec["metadata"]["backend"] == "playwright"
    assert options.capture.enabled is False
    assert options.capture.max_pages == 2
    assert list(options.capture.viewports) == ["mobile"]
    assert options.capture.scroll is False
    assert options.capture.post_nav_wait_s == 1.25
    assert options.capture.max_total_duration_s == 40.0
    assert options.capture.pre_capture_wait_s == 0.5
    assert options.browser.allow_autoplay is False
    assert options.browser.hide_overlays is False
    assert options.browser.compatibility_level == "medium"
    assert options.browser.user_agent == "NestedUA/1.0"
    assert "body { color: green; }" in options.runner.extra_styles
    assert "console.log('nested');" in options.runner.extra_init_scripts
    assert options.runner.extra["locale"] == "fr-FR"
    assert options.runner.playwright_executable_path == "/usr/bin/google-chrome-stable"


def test_collect_job_specs_raises_on_schema_mismatch(tmp_path: Path) -> None:
    input_path = tmp_path / "jobs_invalid_schema.jsonl"
    input_path.write_text(
        json.dumps(
            {
                "job_id": "blog-detail",
                "url": "https://example.com/blog/intro",
                "options": {
                    "schema_version": "screenshot_options/v99",
                    "capture": {"enabled": True},
                },
            }
        )
        + "\n",
        encoding="utf-8",
    )

    cli_args = ScreenshotCliArgs(
        input=input_path,
        urls=(),
        site_ids=(),
        partition_date=None,
        max_pages=5,
        depth=1,
        viewports=("desktop",),
        post_nav_wait_s=0.0,
        timeout_s=60.0,
        max_retries=None,
        job_budget_s=None,
        scroll=False,
        max_viewport_concurrency=None,
    )

    with pytest.raises(ValueError, match="Unsupported screenshot options schema"):
        collect_job_specs(cli_args)


def test_collect_job_specs_boolean_overrides_from_args() -> None:
    cli_args = ScreenshotCliArgs(
        input=None,
        urls=("https://example.com",),
        site_ids=(),
        partition_date=None,
        max_pages=1,
        depth=0,
        viewports=("desktop",),
        post_nav_wait_s=0.0,
        timeout_s=60.0,
        max_retries=None,
        job_budget_s=None,
        allow_autoplay=False,
        hide_overlays=False,
        reduced_motion=True,
        full_page=False,
        pre_capture_wait_s=None,
        max_viewport_concurrency=None,
        override_custom_user_agent="CustomAgent/1.0",
    )

    spec = collect_job_specs(cli_args)[0]
    options = spec["options"]
    assert spec["backend"] == "playwright"
    assert spec["metadata"]["backend"] == "playwright"

    assert options.browser.allow_autoplay is False
    assert options.browser.hide_overlays is False
    assert options.browser.reduced_motion is True
    assert options.capture.full_page is False
    assert options.browser.mute_media is True
    assert options.browser.user_agent == "CustomAgent/1.0"
    assert options.browser.disable_animations is True
    assert options.browser.block_media is False
    assert options.runner.playwright_executable_path is None


def test_collect_job_specs_with_playwright_executable_path(tmp_path: Path) -> None:
    cli_args = ScreenshotCliArgs(
        input=None,
        urls=("https://example.com",),
        site_ids=(),
        partition_date=None,
        max_pages=1,
        depth=0,
        viewports=("desktop",),
        post_nav_wait_s=0.0,
        timeout_s=60.0,
        max_retries=None,
        job_budget_s=None,
        playwright_executable_path="/opt/google/chrome",
    )

    spec = collect_job_specs(cli_args)[0]
    assert spec["options"].runner.playwright_executable_path == "/opt/google/chrome"


def test_collect_job_specs_respects_overrides(tmp_path: Path) -> None:
    input_path = tmp_path / "jobs.jsonl"
    input_path.write_text(
        json.dumps(
            {
                "url": "https://example.com/about",
                "options": {
                    "capture": {
                        "enabled": True,
                        "viewports": ["desktop"],
                        "max_capture_attempts": 5,
                        "max_total_duration_s": 25.0,
                        "max_viewport_concurrency": 2,
                        "full_page": False,
                        "pre_capture_wait_s": 1.5,
                    },
                    "browser": {
                        "allow_autoplay": False,
                        "hide_overlays": False,
                        "reduced_motion": True,
                        "block_media": True,
                        "mute_media": False,
                        "disable_animations": True,
                    },
                    "runner": {
                        "extra_styles": ["body { background: #000; }"],
                        "extra_init_scripts": ["console.log('hi')"],
                        "extra": {"locale": "en-US"},
                    },
                },
                "metadata": ["ignored"],
            }
        )
        + "\n\n"
        + json.dumps({"url": "   "})
        + "\n",
        encoding="utf-8",
    )

    cli_args = ScreenshotCliArgs(
        input=input_path,
        urls=(),
        site_ids=(),
        partition_date="2025-10-23",
        max_pages=5,
        depth=2,
        viewports=("desktop", "mobile"),
        post_nav_wait_s=0.0,
        timeout_s=30.0,
        max_retries=None,
        job_budget_s=120.0,
        max_viewport_concurrency=None,
        allow_autoplay=True,
        hide_overlays=True,
        reduced_motion=False,
        full_page=True,
        pre_capture_wait_s=None,
        chromium_compat="medium",
        override_custom_user_agent="CLI/2.0",
    )

    specs = collect_job_specs(cli_args)
    assert len(specs) == 1
    spec: ScreenshotJobSpec = specs[0]
    options = spec["options"]
    assert spec["backend"] == "playwright"
    assert spec["metadata"]["backend"] == "playwright"
    assert list(options.capture.viewports) == ["desktop"]
    assert options.capture.max_capture_attempts == 5
    assert options.capture.max_total_duration_s == 25.0
    assert options.browser.allow_autoplay is False
    assert options.browser.hide_overlays is False
    assert options.browser.reduced_motion is True
    assert options.capture.full_page is False
    assert options.capture.pre_capture_wait_s == 1.5
    assert options.browser.disable_animations is True
    assert options.browser.block_media is True
    assert options.browser.mute_media is False
    assert options.browser.user_agent == "CLI/2.0"
    assert options.capture.max_viewport_concurrency == 2
    assert "body { background: #000; }" in options.runner.extra_styles
    assert "console.log('hi')" in options.runner.extra_init_scripts
    assert options.runner.extra["locale"] == "en-US"
    assert options.browser.compatibility_level == "medium"
    assert specs[0]["metadata"]["source"] == "cli"


def test_collect_job_specs_uses_arg_level_pre_capture_wait(tmp_path: Path) -> None:
    cli_args = ScreenshotCliArgs(
        input=None,
        urls=("https://example.com/blog",),
        site_ids=(),
        partition_date=None,
        max_pages=2,
        depth=0,
        viewports=("desktop",),
        post_nav_wait_s=0.0,
        timeout_s=40.0,
        max_retries=None,
        job_budget_s=None,
        scroll=False,
        pre_capture_wait_s=3.2,
        max_viewport_concurrency=None,
    )

    specs = collect_job_specs(cli_args)
    spec: ScreenshotJobSpec = specs[0]
    options = spec["options"]
    assert spec["backend"] == "playwright"
    assert spec["metadata"]["backend"] == "playwright"
    assert options.capture.scroll is False
    assert options.capture.pre_capture_wait_s == 3.2
    assert options.browser.mute_media is True
    assert options.browser.user_agent is None


def test_spec_to_api_payload_includes_snapshot() -> None:
    options = ScreenshotOptions(
        capture=CaptureOptions(enabled=True, max_pages=1),
        browser=BrowserCompatOptions(user_agent="SpecUA/1.0"),
    )
    spec: ScreenshotJobSpec = {
        "job_id": "job-1",
        "url": "https://example.com",
        "partition_date": "2025-01-01",
        "metadata": {"source": "cli"},
        "options": options,
        "html_snapshot_path": "/tmp/snapshot.html",
        "backend": "playwright",
    }

    payload: dict[str, object] = spec_to_api_payload(spec)
    options_payload = cast(dict[str, object], payload["options"])
    capture_payload = cast(dict[str, object], options_payload["capture"])
    browser_payload = cast(dict[str, object], options_payload["browser"])
    runner_payload = cast(dict[str, object], options_payload["runner"])

    assert payload["html_snapshot_path"] == "/tmp/snapshot.html"
    assert options_payload["schema_version"] == OPTIONS_SCHEMA_VERSION
    assert capture_payload["enabled"] is True
    assert browser_payload["mute_media"] is True
    assert browser_payload.get("user_agent") == "SpecUA/1.0"
    assert runner_payload.get("playwright_executable_path") is None
    assert payload["backend"] == "playwright"


def test_serialize_options_round_trip() -> None:
    options = ScreenshotOptions(
        capture=CaptureOptions(
            enabled=True,
            max_pages=3,
            depth=2,
            viewports=("desktop", "mobile"),
            post_nav_wait_s=1.0,
            timeout_s=30.0,
            scroll=False,
            max_total_duration_s=120.0,
            max_capture_attempts=4,
            full_page=False,
            pre_capture_wait_s=0.75,
            max_viewport_concurrency=5,
        ),
        browser=BrowserCompatOptions(
            allow_autoplay=False,
            hide_overlays=False,
            reduced_motion=True,
            disable_animations=True,
            mute_media=False,
            block_media=True,
            compatibility_level="high",
            user_agent="UA/3.0",
        ),
        runner=RunnerOptions(
            extra_styles=("body { color: red; }",),
            extra_init_scripts=("console.log('hi')",),
            extra={"foo": "bar"},
        ),
    )

    serialized: dict[str, object] = serialize_options(options)
    capture_payload = cast(dict[str, object], serialized["capture"])
    browser_payload = cast(dict[str, object], serialized["browser"])
    runner_payload = cast(dict[str, object], serialized["runner"])

    assert serialized["schema_version"] == OPTIONS_SCHEMA_VERSION
    assert capture_payload["viewports"] == ["desktop", "mobile"]
    assert capture_payload["max_capture_attempts"] == 4
    assert capture_payload["max_viewport_concurrency"] == 5
    assert browser_payload["mute_media"] is False
    assert browser_payload["disable_animations"] is True
    assert browser_payload["block_media"] is True
    assert browser_payload["compatibility_level"] == "high"
    assert browser_payload["user_agent"] == "UA/3.0"
    assert runner_payload["extra"] == {"foo": "bar"}
    assert "body { color: red; }" in cast(list[str], runner_payload["extra_styles"])
    assert "console.log('hi')" in cast(list[str], runner_payload["extra_init_scripts"])
    assert runner_payload["playwright_executable_path"] is None


def test_collect_job_specs_reads_extra_files(tmp_path: Path) -> None:
    css_path = tmp_path / "extra.css"
    css_path.write_text("body { color: red; }", encoding="utf-8")
    js_path = tmp_path / "extra.js"
    js_path.write_text("console.log('css');", encoding="utf-8")

    cli_args = ScreenshotCliArgs(
        input=None,
        urls=("https://example.com",),
        site_ids=(),
        partition_date=None,
        max_pages=1,
        depth=0,
        viewports=("desktop",),
        post_nav_wait_s=0.0,
        timeout_s=60.0,
        max_retries=None,
        job_budget_s=None,
        scroll=None,
        extra_css_paths=(css_path,),
        extra_js_paths=(js_path,),
        chromium_compat="high",
        max_viewport_concurrency=None,
        override_custom_user_agent="UA/2.0",
    )

    spec = collect_job_specs(cli_args)[0]
    options = spec["options"]
    assert spec["backend"] == "playwright"

    assert "body { color: red; }" in options.runner.extra_styles
    assert "console.log('css');" in options.runner.extra_init_scripts
    assert options.browser.compatibility_level == "high"
    assert options.capture.max_viewport_concurrency == 1
    assert options.browser.user_agent == "UA/2.0"


def test_collect_job_specs_rejects_unsupported_backend() -> None:
    cli_args = ScreenshotCliArgs(
        input=None,
        urls=("https://example.org",),
        site_ids=(),
        partition_date=None,
        max_pages=1,
        depth=0,
        viewports=("desktop",),
        post_nav_wait_s=0.0,
        timeout_s=30.0,
        max_retries=None,
        job_budget_s=None,
        backend="selenium",
    )

    with pytest.raises(ValueError):
        collect_job_specs(cli_args)


def test_derive_job_id_fallback() -> None:
    assert derive_job_id("", fallback="site-1") == "site-1"



"""Shared models and services for screenshot capture backends."""

from __future__ import annotations

__version__ = "0.1.5"

import logging

from ._shared.storage import (
    AsyncStorageBackend,
    CloudStorageBackend,
    LocalStorageBackend,
    StorageBackend,
)
from ._shared.types import ScreenshotJobSpec, ScreenshotMetadata
from .models import (
    ScreenshotBatchResult,
    ScreenshotCaptureResult,
    ScreenshotJob,
    ScreenshotOptions,
    ScreenshotResourceError,
    ScreenshotResourceOutcome,
    ScreenshotResourceResult,
)
from .playwright_runner import KNOWN_VIEWPORTS, capture_screenshots, capture_screenshots_async
from .selenium_runner import capture_screenshots_async as capture_screenshots_async_selenium
from .service import ScreenshotBackend, ScreenshotService

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "ScreenshotBatchResult",
    "ScreenshotCaptureResult",
    "ScreenshotJob",
    "ScreenshotOptions",
    "ScreenshotResourceError",
    "ScreenshotResourceOutcome",
    "ScreenshotResourceResult",
    "ScreenshotJobSpec",
    "ScreenshotMetadata",
    "ScreenshotBackend",
    "ScreenshotService",
    "KNOWN_VIEWPORTS",
    "LocalStorageBackend",
    "StorageBackend",
    "CloudStorageBackend",
    "AsyncStorageBackend",
    "capture_screenshots",
    "capture_screenshots_async",
    "capture_screenshots_async_selenium",
]



"""Centralized defaults and env-driven settings for screenshot package.

Use these helpers to avoid magic numbers scattered across modules and to
support simple environment overrides in CI or operational contexts.
"""

from __future__ import annotations

import os
from dataclasses import dataclass


def _int_env(name: str, default: int) -> int:
    try:
        raw = os.getenv(name)
        return int(raw) if raw is not None else default
    except Exception:
        return default


def _float_env(name: str, default: float) -> float:
    try:
        raw = os.getenv(name)
        return float(raw) if raw is not None else default
    except Exception:
        return default


def _bool_env(name: str, default: bool) -> bool:
    raw = os.getenv(name)
    if raw is None:
        return default
    return str(raw).strip().lower() in {"1", "true", "yes", "on"}


@dataclass(frozen=True)
class ScreenshotDefaults:
    scroll_step_delay_ms: int = 250
    max_scroll_steps: int = 15
    playwright_max_capture_attempts: int = 3
    retry_backoff_s: float = 0.5
    retry_max_backoff_s: float = 5.0
    enable_timing: bool = False

    @classmethod
    def load(cls) -> ScreenshotDefaults:
        return cls(
            scroll_step_delay_ms=_int_env("SCREENSHOT_SCROLL_STEP_DELAY_MS", 250),
            max_scroll_steps=_int_env("SCREENSHOT_MAX_SCROLL_STEPS", 15),
            playwright_max_capture_attempts=_int_env("PLAYWRIGHT_CAPTURE_MAX_ATTEMPTS", 3),
            retry_backoff_s=_float_env("SCREENSHOT_RETRY_BACKOFF_S", 0.5),
            retry_max_backoff_s=_float_env("SCREENSHOT_RETRY_MAX_BACKOFF_S", 5.0),
            enable_timing=_bool_env("SCREENSHOT_ENABLE_TIMING", False),
        )


DEFAULTS = ScreenshotDefaults.load()



"""Programmatic CLI argument builders shared by screenshot entry points."""

from __future__ import annotations

import argparse
from collections.abc import Callable
from pathlib import Path

__all__ = ["add_programmatic_job_arguments", "_parse_bool"]


def _parse_bool(value: str | bool | None) -> bool:
    if value is None:
        return True
    if isinstance(value, bool):
        return value
    normalized = str(value).strip().lower()
    if normalized in {"1", "true", "yes", "y"}:
        return True
    if normalized in {"0", "false", "no", "n"}:
        return False
    raise argparse.ArgumentTypeError(f"expected boolean value, got {value!r}")


def add_programmatic_job_arguments(
    parser: argparse.ArgumentParser,
    *,
    include_output_dir: bool,
    normalize_chromium_compat: Callable[[str], str],
) -> None:
    """Add programmatic job arguments used when config-driven parsing is disabled."""

    parser.add_argument(
        "--urls",
        action="append",
        dest="urls",
        metavar="URL",
        help="Website URL to capture (repeat flag to capture multiple sites)",
    )
    parser.add_argument(
        "--job-id",
        action="append",
        dest="site_ids",
        help="Identifier to use for the corresponding URL (repeatable; defaults to derived name)",
    )
    parser.add_argument(
        "--input",
        type=Path,
        help='Optional JSONL file with objects {"url": ..., "job_id": ...}',
    )
    parser.add_argument(
        "--backend",
        choices=("playwright", "selenium"),
        default="playwright",
        help="Screenshot backend to use (default: playwright).",
    )

    if include_output_dir:
        parser.add_argument(
            "--output-dir",
            type=Path,
            required=True,
            help="Directory where screenshots and metadata will be stored",
        )

    parser.add_argument(
        "--partition-date",
        type=str,
        help="Optional partition date (YYYY-MM-DD) to embed in metadata",
    )
    parser.add_argument(
        "--max-pages",
        dest="max_pages",
        type=int,
        help="Maximum number of pages to capture per site (default: 5)",
    )
    parser.add_argument(
        "--depth",
        dest="depth",
        type=int,
        help="Maximum link depth to follow when capturing screenshots (default: 1)",
    )
    parser.add_argument(
        "--viewports",
        dest="viewports",
        nargs="+",
        help="Viewport presets for screenshot capture (default: desktop)",
    )
    parser.add_argument(
        "--post-nav-wait-s",
        dest="post_nav_wait_s",
        type=float,
        help="Seconds to wait after navigation before scripted settling (default: 6).",
    )
    parser.add_argument(
        "--timeout-s",
        dest="timeout",
        type=float,
        help="Timeout in seconds for each page capture attempt (default: 60).",
    )
    parser.add_argument(
        "--max-retries",
        dest="max_retries",
        type=int,
        help=(
            "Maximum capture retries per viewport (default: 3 or PLAYWRIGHT_CAPTURE_MAX_ATTEMPTS)."
        ),
    )
    parser.add_argument(
        "--job-budget-s",
        dest="job_budget_s",
        type=float,
        help="Overall time budget in seconds per job before following new links.",
    )
    parser.add_argument(
        "--max-viewport-concurrency",
        dest="max_viewport_concurrency",
        type=int,
        help="Maximum number of viewports captured concurrently (default: 3).",
    )
    parser.add_argument(
        "--scroll",
        dest="scroll",
        type=_parse_bool,
        metavar="{true|false}",
        help="Enable or disable pre-capture scrolling (default: true).",
    )
    parser.add_argument(
        "--scroll-step-delay-ms",
        dest="scroll_step_delay_ms",
        type=int,
        help="Delay between scroll steps in milliseconds (default: 250).",
    )
    parser.add_argument(
        "--max-scroll-steps",
        dest="max_scroll_steps",
        type=int,
        help="Maximum scroll iterations per capture (default: 15).",
    )

    parser.add_argument(
        "--allow-autoplay",
        dest="allow_autoplay",
        nargs="?",
        const="true",
        type=_parse_bool,
        help="Control Chromium autoplay allowance (default: true).",
    )

    parser.add_argument(
        "--hide-overlays",
        dest="hide_overlays",
        nargs="?",
        const="true",
        type=_parse_bool,
        help="Control overlay-hiding CSS injection (default: true).",
    )

    parser.add_argument(
        "--reduced-motion",
        dest="reduced_motion",
        nargs="?",
        const="true",
        type=_parse_bool,
        help="Force prefers-reduced-motion media emulation (default: true).",
    )

    parser.add_argument(
        "--full-page",
        dest="full_page",
        nargs="?",
        const="true",
        type=_parse_bool,
        help="Capture full-page screenshots (default: true). Pass false for viewport-only.",
    )

    parser.add_argument(
        "--mute-media",
        dest="mute_media",
        nargs="?",
        const="true",
        type=_parse_bool,
        help="Force media elements muted during capture (default: true).",
    )

    parser.add_argument(
        "--disable-animations",
        dest="disable_animations",
        nargs="?",
        const="true",
        type=_parse_bool,
        help="Disable CSS animations and transitions before capture (default: true).",
    )

    parser.add_argument(
        "--block-media",
        dest="block_media",
        nargs="?",
        const="true",
        type=_parse_bool,
        help="Block heavyweight media requests such as video/audio (default: false).",
    )

    parser.add_argument(
        "--pre-capture-wait-s",
        dest="pre_capture_wait_s",
        type=float,
        help="Seconds to wait after scrolling/settling before capturing (default: 2.5).",
    )
    parser.add_argument(
        "--extra-css",
        dest="extra_css",
        type=Path,
        action="append",
        help="Path to CSS snippet injected before capture (repeatable).",
    )
    parser.add_argument(
        "--extra-js",
        dest="extra_js",
        type=Path,
        action="append",
        help="Path to JS snippet executed before page scripts (repeatable).",
    )
    parser.add_argument(
        "--override-custom-user-agent",
        dest="override_custom_user_agent",
        type=str,
        help="Override the browser User-Agent string that Playwright would normally supply.",
    )
    parser.add_argument(
        "--playwright-executable-path",
        dest="playwright_executable_path",
        type=str,
        help="Override the Playwright Chromium executable path.",
    )
    parser.add_argument(
        "--chromium-compat",
        dest="chromium_compat",
        type=normalize_chromium_compat,
        help="Chromium compatibility profile (default: high).",
    )

    parser.set_defaults(
        max_pages=None,
        depth=None,
        viewports=None,
        post_nav_wait_s=None,
        timeout=None,
        max_retries=None,
        job_budget_s=None,
        scroll=None,
        allow_autoplay=None,
        hide_overlays=None,
        reduced_motion=None,
        full_page=None,
        mute_media=None,
        disable_animations=None,
        block_media=None,
        pre_capture_wait_s=None,
        extra_css=None,
        extra_js=None,
        chromium_compat=None,
        max_viewport_concurrency=None,
        override_custom_user_agent=None,
        playwright_executable_path=None,
        backend="playwright",
    )



"""Transformations moved out of the public dataclasses for easier testing."""

from __future__ import annotations

import json
from hashlib import sha256

from .._models_results import (
    ScreenshotCaptureResult,
    ScreenshotResourceError,
    ScreenshotResourceOutcome,
    ScreenshotResourceResult,
    ScreenshotViewportResource,
)
from .._shared.errors import ErrorCategory, ScreenshotError


def _entry_succeeded(entry: dict[str, object]) -> bool:
    return str(entry.get("status") or "").lower() == "success" and not entry.get("error")


def _entry_failed(entry: dict[str, object]) -> bool:
    return not _entry_succeeded(entry)


def _determine_viewport_outcome(
    *,
    global_outcome: ScreenshotResourceOutcome,
    requested: int,
    captured: int,
    failed: int,
) -> ScreenshotResourceOutcome:
    if requested == 0:
        return global_outcome
    if failed == 0:
        return ScreenshotResourceOutcome.SUCCESS
    if captured == 0:
        return ScreenshotResourceOutcome.FAILED
    return ScreenshotResourceOutcome.PARTIAL


def _categorize_screenshot_error(err: ScreenshotError) -> str:
    return ErrorCategory.from_error_type(err.error_type).value


def generate_screenshot_config_key(config: dict[str, object]) -> str:
    viewports_val = config.get("viewports", [])
    if isinstance(viewports_val, list | tuple):
        viewports_list = [str(v) for v in viewports_val]
    else:
        viewports_list = [str(viewports_val)] if viewports_val is not None else []
    normalized = {
        "viewports": sorted(viewports_list),
        "depth": config.get("depth", 0),
        "timeout_s": config.get("timeout_s", 60.0),
        "scroll": bool(config.get("scroll", False)),
        "full_page": bool(config.get("full_page", False)),
    }
    payload = json.dumps(normalized, sort_keys=True, separators=(",", ":"))
    return sha256(payload.encode("utf-8")).hexdigest()


def from_capture_result(
    capture_result: ScreenshotCaptureResult,
    *,
    resource_key: str,
    config: dict[str, object],
    metadata: dict[str, object] | None = None,
) -> ScreenshotResourceResult:
    if capture_result.requested == 0:
        outcome = ScreenshotResourceOutcome.SKIPPED
    elif capture_result.captured == 0 and capture_result.requested > 0:
        outcome = ScreenshotResourceOutcome.FAILED
    elif capture_result.failed > 0:
        outcome = ScreenshotResourceOutcome.PARTIAL
    else:
        outcome = ScreenshotResourceOutcome.SUCCESS

    resource_errors = [
        ScreenshotResourceError(
            category=_categorize_screenshot_error(err),
            message=err.message,
            retryable=err.retryable,
            details=dict(err.details),
        )
        for err in capture_result.errors
    ]

    default_metadata = {
        "requested": capture_result.requested,
        "captured": capture_result.captured,
        "failed": capture_result.failed,
        "job_id": capture_result.job_id,
        "metadata_path": str(capture_result.metadata_path)
        if capture_result.metadata_path
        else None,
        "entries": list(capture_result.entries),
    }

    return ScreenshotResourceResult(
        resource_key=resource_key,
        outcome=outcome,
        config_key=generate_screenshot_config_key(config),
        errors=resource_errors,
        metadata={**default_metadata, **(metadata or {})},
        entries=list(capture_result.entries),
    )


def per_viewport_resources(
    resource_result: ScreenshotResourceResult,
) -> list[ScreenshotViewportResource]:
    if not resource_result.entries:
        return [
            ScreenshotViewportResource(
                resource_key=resource_result.resource_key,
                viewport="default",
                outcome=resource_result.outcome,
                config_key=resource_result.config_key,
                errors=list(resource_result.errors),
                metadata=dict(resource_result.metadata),
            )
        ]

    by_viewport: dict[str, list[dict[str, object]]] = {}
    for entry in resource_result.entries:
        viewport = str(entry.get("viewport") or "default").lower()
        by_viewport.setdefault(viewport, []).append(entry)

    results: list[ScreenshotViewportResource] = []
    for viewport, entries in by_viewport.items():
        requested = len(entries)
        captured = sum(1 for e in entries if _entry_succeeded(e))
        failed = sum(1 for e in entries if _entry_failed(e))
        metadata = dict(resource_result.metadata)
        metadata.update(
            {
                "viewport": viewport,
                "entries": entries,
                "requested": requested,
                "captured": captured,
                "failed": failed,
            }
        )
        viewport_outcome = _determine_viewport_outcome(
            global_outcome=resource_result.outcome,
            requested=requested,
            captured=captured,
            failed=failed,
        )
        results.append(
            ScreenshotViewportResource(
                resource_key=resource_result.resource_key,
                viewport=viewport,
                outcome=viewport_outcome,
                config_key=resource_result.config_key,
                errors=list(resource_result.errors),
                metadata=metadata,
            )
        )
    return results



import argparse
import json
from pathlib import Path

import pytest

from screenshot import (
    ScreenshotBatchResult,
    ScreenshotCaptureResult,
    ScreenshotJob,
    ScreenshotOptions,
)
from screenshot._models_options import OPTIONS_SCHEMA_VERSION, CaptureOptions
from screenshot._shared.errors import make_error
from screenshot.cli import (
    PlaywrightScreenshotBackend,
    SeleniumScreenshotBackend,
    _cmd_local,
    build_local_screenshot_service,
    main as cli_main,
)


class DummyService:
    def __init__(self, result: ScreenshotBatchResult) -> None:
        self._result = result
        self.calls = 0
        self.jobs = None

    def capture(self, jobs):
        self.calls += 1
        self.jobs = list(jobs)
        return self._result


def _build_args(tmp_path: Path, **overrides) -> argparse.Namespace:
    defaults = {
        "input": None,
        "urls": (),
        "site_ids": (),
        "partition_date": None,
        "max_pages": 5,
        "depth": 1,
        "viewports": ("desktop",),
        "post_nav_wait_s": 0.0,
        "timeout_s": 60.0,
        "max_retries": None,
        "job_budget_s": None,
        "scroll": None,
        "allow_autoplay": None,
        "hide_overlays": None,
        "reduced_motion": None,
        "full_page": None,
        "pre_capture_wait_s": None,
        "mute_media": None,
        "disable_animations": None,
        "block_media": None,
        "chromium_compat": None,
        "extra_css": None,
        "extra_js": None,
        "output_dir": tmp_path / "output",
        "site_concurrency": 1,
        "summary_file": None,
        "backend": "playwright",
        "playwright_executable_path": None,
    }
    defaults.update(overrides)
    return argparse.Namespace(**defaults)


def test_cmd_local_writes_summary_and_calls_service(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    spec = {
        "job_id": "marketing-home",
        "url": "https://example.com",
        "partition_date": "2024-11-01",
        "html_snapshot_path": None,
        "metadata": {"source": "cli", "backend": "playwright"},
        "options": ScreenshotOptions(capture=CaptureOptions(enabled=True, max_pages=2)),
        "backend": "playwright",
    }
    batch_result = ScreenshotBatchResult(
        results=[
            ScreenshotCaptureResult(
                requested=1,
                captured=1,
                failed=0,
                metadata_path=None,
                entries=[{"job_id": "marketing-home"}],
                errors=[],
                job_id="marketing-home",
            )
        ]
    )
    dummy_service = DummyService(batch_result)
    summary_path = tmp_path / "summary.json"

    monkeypatch.setattr("screenshot.cli.collect_job_specs", lambda _: [spec])

    def _build_service(**kwargs):
        assert kwargs.get("backend") == "playwright"
        return dummy_service

    monkeypatch.setattr("screenshot.cli.build_local_screenshot_service", _build_service)

    args = _build_args(tmp_path, summary_file=summary_path)

    exit_code = _cmd_local(args)

    assert exit_code == 0
    assert dummy_service.calls == 1
    assert dummy_service.jobs and dummy_service.jobs[0].job_id == "marketing-home"
    assert summary_path.exists()
    summary = json.loads(summary_path.read_text(encoding="utf-8"))
    assert summary["success_count"] == 1
    assert summary["failure_count"] == 0


def test_cmd_local_handles_missing_jobs(
    monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture, tmp_path: Path
) -> None:
    caplog.set_level("ERROR")
    monkeypatch.setattr("screenshot.cli.collect_job_specs", lambda _: [])
    monkeypatch.setattr(
        "screenshot.cli.build_local_screenshot_service",
        lambda **_: pytest.fail("should not build service"),
    )

    args = _build_args(tmp_path)

    exit_code = _cmd_local(args)

    assert exit_code == 1
    assert "No screenshot jobs specified" in caplog.text


def test_cmd_local_file_not_found(
    monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture, tmp_path: Path
) -> None:
    caplog.set_level("ERROR")

    def raise_not_found(_):
        raise FileNotFoundError("jobs.jsonl")

    monkeypatch.setattr("screenshot.cli.collect_job_specs", raise_not_found)
    monkeypatch.setattr(
        "screenshot.cli.build_local_screenshot_service",
        lambda **_: pytest.fail("should not build service"),
    )

    args = _build_args(tmp_path, input=tmp_path / "missing.jsonl")

    exit_code = _cmd_local(args)

    assert exit_code == 1
    assert "Input file" in caplog.text


def test_cmd_local_json_error(
    monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture, tmp_path: Path
) -> None:
    caplog.set_level("ERROR")

    def raise_json_error(_):
        raise json.JSONDecodeError("bad json", "{}", 0)

    monkeypatch.setattr("screenshot.cli.collect_job_specs", raise_json_error)
    monkeypatch.setattr(
        "screenshot.cli.build_local_screenshot_service",
        lambda **_: pytest.fail("should not build service"),
    )

    args = _build_args(tmp_path, input=tmp_path / "invalid.jsonl")

    exit_code = _cmd_local(args)

    assert exit_code == 1
    assert "Failed to parse" in caplog.text


def test_cmd_local_logs_failed_jobs(
    monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture, tmp_path: Path
) -> None:
    caplog.set_level("INFO")
    spec = {
        "job_id": "marketing-home",
        "url": "https://example.com",
        "partition_date": None,
        "html_snapshot_path": None,
        "metadata": {"source": "cli", "backend": "playwright"},
        "options": ScreenshotOptions(capture=CaptureOptions(enabled=True)),
        "backend": "playwright",
    }
    batch_result = ScreenshotBatchResult(
        results=[
            ScreenshotCaptureResult(
                requested=1,
                captured=0,
                failed=1,
                metadata_path=None,
                entries=[],
                errors=[make_error("runtime", "boom")],
                job_id="marketing-home",
            )
        ]
    )
    dummy_service = DummyService(batch_result)

    monkeypatch.setattr("screenshot.cli.collect_job_specs", lambda _: [spec])
    monkeypatch.setattr("screenshot.cli.build_local_screenshot_service", lambda **_: dummy_service)

    args = _build_args(tmp_path)

    exit_code = _cmd_local(args)

    assert exit_code == 1
    assert "failed: boom" in caplog.text


def test_cli_local_integration_with_json_input(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    input_path = tmp_path / "jobs.jsonl"
    input_path.write_text(
        json.dumps(
            {
                "job_id": "integration-case",
                "url": "https://example.com",
                "partition_date": "2025-12-01",
                "options": {
                    "schema_version": OPTIONS_SCHEMA_VERSION,
                    "capture": {
                        "enabled": True,
                        "max_pages": 2,
                        "viewports": ["desktop", "mobile"],
                    },
                    "browser": {"allow_autoplay": False, "compatibility_level": "medium"},
                },
                "metadata": {"source": "jsonl"},
            }
        )
        + "\n",
        encoding="utf-8",
    )

    summary_path = tmp_path / "summary.json"
    output_dir = tmp_path / "output"

    metadata_path = output_dir / "integration-case" / "metadata" / "screenshots.json"
    metadata_path.parent.mkdir(parents=True, exist_ok=True)
    metadata_path.write_text(
        json.dumps({"entries": [{"screenshot_path": "shot.png", "status": "success"}]}),
        encoding="utf-8",
    )

    capture_result = ScreenshotCaptureResult(
        requested=1,
        captured=1,
        failed=0,
        metadata_path=metadata_path,
        entries=[
            {
                "job_id": "integration-case",
                "url": "https://example.com",
                "status": "success",
                "screenshot_path": str(
                    metadata_path.parent.parent / "assets" / "screenshots" / "shot.png"
                ),
                "error": None,
            }
        ],
        errors=[],
        job_id="integration-case",
    )
    batch_result = ScreenshotBatchResult(results=[capture_result])

    class StubService:
        def __init__(self, result: ScreenshotBatchResult) -> None:
            self.result = result
            self.calls = 0
            self.jobs: list[ScreenshotJob] | None = None

        def capture(self, jobs):
            self.calls += 1
            self.jobs = list(jobs)
            return self.result

    stub_service = StubService(batch_result)
    monkeypatch.setattr("screenshot.cli.build_local_screenshot_service", lambda **_: stub_service)

    exit_code = cli_main(
        [
            "local",
            "--input",
            str(input_path),
            "--output-dir",
            str(output_dir),
            "--summary-file",
            str(summary_path),
            "--backend",
            "playwright",
        ]
    )

    assert exit_code == 0
    assert stub_service.calls == 1
    assert stub_service.jobs is not None
    job = stub_service.jobs[0]
    assert job.job_id == "integration-case"
    assert list(job.options.capture.viewports) == ["desktop", "mobile"]
    assert job.options.capture.max_pages == 2
    assert summary_path.exists()
    summary_payload = json.loads(summary_path.read_text(encoding="utf-8"))
    assert summary_payload["success_count"] == 1
    assert summary_payload["failure_count"] == 0


@pytest.mark.asyncio
async def test_playwright_backend_runs_job(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    captured = {}

    async def fake_capture_async(*args, **kwargs):
        captured["args"] = args
        captured["kwargs"] = kwargs
        return ScreenshotCaptureResult(
            requested=1,
            captured=1,
            failed=0,
            metadata_path=None,
            entries=[],
            errors=[],
            job_id=None,
        )

    monkeypatch.setattr("screenshot.cli.playwright_capture_async", fake_capture_async)
    storage = object()
    backend = PlaywrightScreenshotBackend(storage=storage)
    job = ScreenshotJob(
        job_id="job-1",
        url="https://example.com",
        output_root=tmp_path,
        options=ScreenshotOptions(capture=CaptureOptions(enabled=True)),
    )

    result = await backend.run_job_async(job)

    assert captured["kwargs"]["storage"] is storage
    assert result.job_id == "job-1"


def test_build_local_screenshot_service_clamps_concurrency() -> None:
    service = build_local_screenshot_service(concurrency=0)
    assert service._default_concurrency == 1


def test_build_local_screenshot_service_selects_backends() -> None:
    service_playwright = build_local_screenshot_service(concurrency=1, backend="playwright")
    assert isinstance(service_playwright._backend, PlaywrightScreenshotBackend)

    service_selenium = build_local_screenshot_service(concurrency=1, backend="selenium")
    assert isinstance(service_selenium._backend, SeleniumScreenshotBackend)


def test_cmd_local_rejects_mixed_backends(
    monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture, tmp_path: Path
) -> None:
    caplog.set_level("ERROR")
    spec_a = {
        "job_id": "site-a",
        "url": "https://example.com",
        "partition_date": None,
        "html_snapshot_path": None,
        "metadata": {"source": "cli", "backend": "playwright"},
        "options": ScreenshotOptions(capture=CaptureOptions(enabled=True)),
        "backend": "playwright",
    }
    spec_b = {
        "job_id": "site-b",
        "url": "https://example.org",
        "partition_date": None,
        "html_snapshot_path": None,
        "metadata": {"source": "cli", "backend": "selenium"},
        "options": ScreenshotOptions(capture=CaptureOptions(enabled=True)),
        "backend": "selenium",
    }

    monkeypatch.setattr("screenshot.cli.collect_job_specs", lambda _: [spec_a, spec_b])

    args = _build_args(tmp_path)
    exit_code = _cmd_local(args)
    assert exit_code == 1
    assert "Mixed backend selections" in caplog.text


def test_main_invokes_cmd_local(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    called = {}

    def fake_cmd_local(args: argparse.Namespace) -> int:
        called["args"] = args
        return 0

    monkeypatch.setattr("screenshot.cli._cmd_local", fake_cmd_local)

    exit_code = cli_main(
        [
            "--verbose",
            "local",
            "--output-dir",
            str(tmp_path / "output"),
            "--urls",
            "https://example.com",
        ]
    )

    assert exit_code == 0
    assert called["args"].verbose is True


def test_main_unknown_command_exits() -> None:
    with pytest.raises(SystemExit):
        cli_main(["unknown"])



from __future__ import annotations

import asyncio
from pathlib import Path

import pytest

from screenshot import (
    ScreenshotBatchResult,
    ScreenshotCaptureResult,
    ScreenshotJob,
    ScreenshotOptions,
    ScreenshotService,
)
from screenshot._shared.errors import make_error
from screenshot.models import CaptureOptions


class RecordingBackend:
    def __init__(self) -> None:
        self.jobs: list[ScreenshotJob] = []

    async def run_job_async(self, job: ScreenshotJob) -> ScreenshotCaptureResult:
        self.jobs.append(job)
        return ScreenshotCaptureResult(
            requested=1,
            captured=1,
            failed=0,
            metadata_path=None,
            entries=[],
            errors=[],
            job_id=job.job_id,
        )


@pytest.mark.asyncio
async def test_screenshot_service_capture_async(tmp_path: Path) -> None:
    backend = RecordingBackend()
    service = ScreenshotService(backend, concurrency=2)
    jobs = [
        ScreenshotJob(
            job_id=f"job-{index}",
            url=f"https://example.com/{index}",
            output_root=tmp_path,
            options=ScreenshotOptions(capture=CaptureOptions(enabled=True)),
        )
        for index in range(3)
    ]

    result = await service.capture_async(jobs)

    assert isinstance(result, ScreenshotBatchResult)
    assert len(result.results) == 3
    assert len(result.failures) == 0
    assert [job.job_id for job in backend.jobs] == [job.job_id for job in jobs]


def test_screenshot_service_capture_sync(tmp_path: Path) -> None:
    backend = RecordingBackend()
    service = ScreenshotService(backend, concurrency=1)
    job = ScreenshotJob(
        job_id="sync",
        url="https://example.com",
        output_root=tmp_path,
        options=ScreenshotOptions(capture=CaptureOptions(enabled=True)),
    )

    result = service.capture([job])

    assert len(result.results) == 1
    assert result.results[0].job_id == "sync"
    assert backend.jobs[0].job_id == "sync"


def test_screenshot_service_requires_positive_concurrency() -> None:
    backend = RecordingBackend()
    with pytest.raises(ValueError):
        ScreenshotService(backend, concurrency=0)


class NullJobIdBackend:
    def __init__(self) -> None:
        self.jobs: list[ScreenshotJob] = []

    async def run_job_async(self, job: ScreenshotJob) -> ScreenshotCaptureResult:
        self.jobs.append(job)
        return ScreenshotCaptureResult(
            requested=1,
            captured=0,
            failed=1,
            metadata_path=None,
            entries=[],
            errors=[make_error("runtime", "failed")],
            job_id=None,
        )


@pytest.mark.asyncio
async def test_screenshot_service_backfills_job_id(tmp_path: Path) -> None:
    backend = NullJobIdBackend()
    service = ScreenshotService(backend, concurrency=2)
    job = ScreenshotJob(
        job_id="job-123",
        url="https://example.com",
        output_root=tmp_path,
        options=ScreenshotOptions(capture=CaptureOptions(enabled=True)),
    )

    result = await service.capture_async([job])

    assert result.results[0].job_id == "job-123"
    assert backend.jobs[0].job_id == "job-123"


@pytest.mark.asyncio
async def test_screenshot_service_checks_cancel_token(tmp_path: Path) -> None:
    class CountingToken:
        def __init__(self) -> None:
            self.calls = 0

        def raise_if_cancelled(self) -> None:
            self.calls += 1

    token = CountingToken()
    backend = RecordingBackend()
    service = ScreenshotService(backend, concurrency=1)
    jobs = [
        ScreenshotJob(
            job_id=f"job-{index}",
            url=f"https://example.com/{index}",
            output_root=tmp_path,
            options=ScreenshotOptions(capture=CaptureOptions(enabled=True)),
        )
        for index in range(2)
    ]

    await service.capture_async(jobs, cancel_token=token)

    assert token.calls == len(jobs)


@pytest.mark.asyncio
async def test_screenshot_service_handles_no_jobs() -> None:
    backend = RecordingBackend()
    service = ScreenshotService(backend, concurrency=1)

    result = await service.capture_async([])

    assert result.results == []


class FailingBackend:
    async def run_job_async(self, job: ScreenshotJob) -> ScreenshotCaptureResult:
        raise RuntimeError(f"boom-{job.job_id}")


@pytest.mark.asyncio
async def test_screenshot_service_converts_backend_error(tmp_path: Path) -> None:
    backend = FailingBackend()
    service = ScreenshotService(backend, concurrency=1)
    job = ScreenshotJob(
        job_id="err",
        url="https://example.com",
        output_root=tmp_path,
        options=ScreenshotOptions(capture=CaptureOptions(enabled=True)),
    )

    result = await service.capture_async([job])

    assert len(result.results) == 1
    failure = result.results[0]
    assert failure.job_id == "err"
    assert failure.failed == 1
    assert failure.errors and "boom-err" in failure.errors[0].message


class CancelledBackend:
    async def run_job_async(
        self, job: ScreenshotJob
    ) -> ScreenshotCaptureResult:  # pragma: no cover
        raise asyncio.CancelledError()


@pytest.mark.asyncio
async def test_screenshot_service_propagates_cancel(tmp_path: Path) -> None:
    backend = CancelledBackend()
    service = ScreenshotService(backend, concurrency=1)
    job = ScreenshotJob(
        job_id="cancel",
        url="https://example.com",
        output_root=tmp_path,
        options=ScreenshotOptions(capture=CaptureOptions(enabled=True)),
    )

    with pytest.raises(asyncio.CancelledError):
        await service.capture_async([job])



"""Shared test utilities for screenshot tests."""

from __future__ import annotations

from pathlib import Path
from typing import Any

from screenshot import ScreenshotCaptureResult
from screenshot._shared.errors import ScreenshotError


def create_capture_result(
    *,
    requested: int,
    captured: int,
    failed: int,
    errors: list[ScreenshotError] | None = None,
    metadata_path: Path | None = None,
    entries: list[dict[str, Any]] | None = None,
    job_id: str = "job-123",
) -> ScreenshotCaptureResult:
    """Create a ScreenshotCaptureResult for testing.

    Args:
        requested: Number of screenshots requested.
        captured: Number of screenshots captured.
        failed: Number of screenshots failed.
        errors: Optional list of errors.
        metadata_path: Path to metadata file (defaults to /tmp/meta.json).
        entries: Metadata entries (defaults to single desktop viewport).
        job_id: Job identifier (defaults to "job-123").

    Returns:
        ScreenshotCaptureResult instance for testing.
    """
    return ScreenshotCaptureResult(
        requested=requested,
        captured=captured,
        failed=failed,
        metadata_path=metadata_path or Path("/tmp/meta.json"),
        entries=entries or [{"viewport": "desktop"}],
        errors=list(errors or []),
        job_id=job_id,
    )



"""Typed builders for nested CLI options -> ScreenshotOptions.

Purpose:
- Provide pure functions that assemble `ScreenshotOptions` from already
  validated, typed inputs (nested "v2" schema and CLI arg snapshots).
- Keep `options.py` thin and focused on orchestration/dispatch.

Notes:
- Helpers here (e.g., `ensure_list`, `pick`, `to_float`) are intentionally
  small and side‑effect free to make unit testing straightforward.
- Public entry point: `build_from_nested()`; it returns a fully populated
  `ScreenshotOptions` without mutating its arguments.
"""

from __future__ import annotations

from collections.abc import Callable, Mapping, Sequence
from typing import Any

from ..._models_options import BrowserCompatOptions, CaptureOptions, RunnerOptions
from ...models import ScreenshotOptions
from .schema import ScreenshotCliArgs

_CHROMIUM_COMPAT_CHOICES = {"low", "medium", "high"}


def ensure_list(value: object | None) -> list[str]:
    if value is None:
        return []
    if isinstance(value, list | tuple | set):
        return [str(item) for item in value if isinstance(item, str)]
    return [str(value)]


def pick(containers: Sequence[Mapping[str, object]], *keys: str) -> object | None:
    for key in keys:
        for container in containers:
            if key in container and container[key] is not None:
                return container[key]
    return None


def to_float(value: object | None, default: float) -> float:
    if value is None:
        return default
    if isinstance(value, int | float):
        return float(value)
    if isinstance(value, str):
        try:
            return float(value.strip())
        except ValueError:
            return default
    return default


def bool_field(
    containers: Sequence[Mapping[str, object]],
    keys: Sequence[str],
    arg_value: bool | None,
    default: bool,
) -> bool:
    raw = pick(containers, *keys)
    if raw is not None:
        return bool(raw)
    if arg_value is not None:
        return bool(arg_value)
    return default


def numeric_field(
    containers: Sequence[Mapping[str, object]],
    keys: Sequence[str],
    *,
    fallback: object | None,
    default: float,
    coerce: Callable[[Any], float],
    minimum: float | None = None,
) -> float:
    raw = pick(containers, *keys)
    if raw is None:
        raw = fallback
    if raw is None:
        raw = default
    value = coerce(raw)
    if minimum is not None:
        value = max(minimum, value)
    return value


def build_from_nested(
    data: Mapping[str, object],
    defaults: ScreenshotOptions,
    args: ScreenshotCliArgs,
    css_snippets: Sequence[str],
    js_snippets: Sequence[str],
) -> ScreenshotOptions:
    capture_data = data.get("capture")
    browser_data = data.get("browser")
    runner_data = data.get("runner")
    capture_data = capture_data if isinstance(capture_data, dict) else {}
    browser_data = browser_data if isinstance(browser_data, dict) else {}
    runner_data = runner_data if isinstance(runner_data, dict) else {}

    max_pages = int(
        numeric_field(
            [capture_data, data],
            ("max_pages",),
            fallback=args.max_pages,
            default=defaults.capture.max_pages,
            coerce=float,
            minimum=1,
        )
    )
    depth = int(
        numeric_field(
            [capture_data, data],
            ("depth",),
            fallback=args.depth,
            default=defaults.capture.depth,
            coerce=float,
            minimum=0,
        )
    )

    viewports_value = pick([capture_data, data], "viewports")
    if viewports_value is None:
        viewports_value = args.viewports or ["desktop"]
    viewports_list = ensure_list(viewports_value)
    if not viewports_list:
        viewports_list = ["desktop"]

    raw_post_nav_wait = pick([capture_data, data], "post_nav_wait_s", "delay")
    post_nav_wait = numeric_field(
        [],
        (),
        fallback=raw_post_nav_wait if raw_post_nav_wait is not None else args.post_nav_wait_s,
        default=defaults.capture.post_nav_wait_s,
        coerce=float,
        minimum=0.0,
    )

    timeout = numeric_field(
        [capture_data, data],
        ("timeout_s", "timeout"),
        fallback=args.timeout_s,
        default=defaults.capture.timeout_s,
        coerce=float,
        minimum=10.0,
    )

    scroll = bool_field([capture_data, data], ("scroll",), args.scroll, defaults.capture.scroll)

    raw_scroll_delay = pick([capture_data, data], "scroll_step_delay_ms", "scroll_step_delay")
    if raw_scroll_delay is None:
        raw_scroll_delay = args.scroll_step_delay_ms
    if raw_scroll_delay is not None:
        scroll_step_delay_ms = int(
            numeric_field(
                [],
                (),
                fallback=raw_scroll_delay,
                default=defaults.capture.scroll_step_delay_ms,
                coerce=float,
                minimum=0.0,
            )
        )
    else:
        scroll_step_delay_ms = defaults.capture.scroll_step_delay_ms

    raw_scroll_steps = pick([capture_data, data], "max_scroll_steps", "scroll_steps")
    if raw_scroll_steps is None:
        raw_scroll_steps = args.max_scroll_steps
    if raw_scroll_steps is not None:
        max_scroll_steps = int(
            numeric_field(
                [],
                (),
                fallback=raw_scroll_steps,
                default=defaults.capture.max_scroll_steps,
                coerce=float,
                minimum=0.0,
            )
        )
    else:
        max_scroll_steps = defaults.capture.max_scroll_steps

    raw_retries = pick([capture_data, data], "max_capture_attempts", "max_retries")
    if raw_retries is None:
        raw_retries = args.max_retries
    if raw_retries is not None:
        max_attempts = int(
            numeric_field(
                [],
                (),
                fallback=raw_retries,
                default=defaults.capture.max_capture_attempts,
                coerce=float,
                minimum=1,
            )
        )
    else:
        max_attempts = defaults.capture.max_capture_attempts

    raw_budget = pick(
        [capture_data, data], "max_total_duration_s", "max_total_duration", "job_budget_s"
    )
    if raw_budget is None:
        raw_budget = args.job_budget_s
    if raw_budget is not None:
        max_total_duration = numeric_field(
            [],
            (),
            fallback=raw_budget,
            default=defaults.capture.max_total_duration_s or 0.0,
            coerce=float,
            minimum=10.0,
        )
    else:
        max_total_duration = None

    raw_viewport_concurrency = pick([capture_data, data], "max_viewport_concurrency")
    if raw_viewport_concurrency is None:
        raw_viewport_concurrency = args.max_viewport_concurrency
    if raw_viewport_concurrency is None:
        max_viewport_concurrency = defaults.capture.max_viewport_concurrency
    else:
        max_viewport_concurrency = int(
            numeric_field(
                [],
                (),
                fallback=raw_viewport_concurrency,
                default=defaults.capture.max_viewport_concurrency,
                coerce=float,
                minimum=1,
            )
        )

    raw_pre_capture = pick([capture_data, data], "pre_capture_wait_s", "settle_timeout")
    if raw_pre_capture is None:
        raw_pre_capture_ms = pick([capture_data, data], "pre_capture_wait_ms")
        if raw_pre_capture_ms is not None:
            raw_pre_capture = to_float(raw_pre_capture_ms, 0.0) / 1000.0
    if raw_pre_capture is None:
        raw_pre_capture = args.pre_capture_wait_s
    settle_timeout = numeric_field(
        [],
        (),
        fallback=raw_pre_capture,
        default=defaults.capture.pre_capture_wait_s,
        coerce=float,
        minimum=0.0,
    )

    full_page = bool_field(
        [capture_data, data], ("full_page",), args.full_page, defaults.capture.full_page
    )

    allow_autoplay = bool_field(
        [browser_data, data],
        ("allow_autoplay",),
        args.allow_autoplay,
        defaults.browser.allow_autoplay,
    )
    hide_overlays = bool_field(
        [browser_data, data],
        ("hide_overlays",),
        args.hide_overlays,
        defaults.browser.hide_overlays,
    )
    reduced_motion = bool_field(
        [browser_data, data],
        ("reduced_motion",),
        args.reduced_motion,
        defaults.browser.reduced_motion,
    )
    mute_media = bool_field(
        [browser_data, data], ("mute_media",), args.mute_media, defaults.browser.mute_media
    )
    disable_animations = bool_field(
        [browser_data, data],
        ("disable_animations",),
        args.disable_animations,
        defaults.browser.disable_animations,
    )
    block_media = bool_field(
        [browser_data, data],
        ("block_media",),
        args.block_media,
        defaults.browser.block_media,
    )

    raw_compat = pick(
        [browser_data, data], "compatibility_level", "chromium_compat", "forgiveness_level"
    )
    if raw_compat is None:
        raw_compat = args.chromium_compat
    if raw_compat is None:
        raw_compat = defaults.browser.compatibility_level
    forgiveness_level = str(raw_compat).strip().lower()
    if forgiveness_level not in _CHROMIUM_COMPAT_CHOICES:
        forgiveness_level = defaults.browser.compatibility_level

    override_user_agent_raw = (
        pick([browser_data], "user_agent")
        or data.get("override_custom_user_agent")
        or data.get("user_agent")
        or args.override_custom_user_agent
        or defaults.browser.user_agent
    )
    override_user_agent = (
        str(override_user_agent_raw) if override_user_agent_raw is not None else None
    )

    extra_styles: list[str] = list(css_snippets)
    runner_styles = pick([runner_data, data], "extra_styles")
    extra_styles.extend(ensure_list(runner_styles))

    extra_init_scripts: list[str] = list(js_snippets)
    runner_scripts = pick([runner_data, data], "extra_init_scripts")
    extra_init_scripts.extend(ensure_list(runner_scripts))

    extra: dict[str, object] = {}
    raw_extra = pick([runner_data, data], "extra")
    if isinstance(raw_extra, dict):
        extra.update(raw_extra)

    playwright_executable_path_raw = pick([runner_data, data], "playwright_executable_path")
    playwright_executable_path = (
        str(playwright_executable_path_raw) if playwright_executable_path_raw is not None else None
    )
    if playwright_executable_path is None:
        playwright_executable_path = args.playwright_executable_path

    capture_opts = CaptureOptions(
        enabled=bool(
            pick([capture_data, data], "enabled")
            if pick([capture_data, data], "enabled") is not None
            else True
        ),
        max_pages=max_pages,
        depth=depth,
        viewports=tuple(viewports_list or ["desktop"]),
        post_nav_wait_s=post_nav_wait,
        timeout_s=timeout,
        max_total_duration_s=max_total_duration,
        max_capture_attempts=max_attempts,
        max_viewport_concurrency=max_viewport_concurrency,
        scroll=scroll,
        scroll_step_delay_ms=scroll_step_delay_ms,
        max_scroll_steps=max_scroll_steps,
        full_page=full_page,
        pre_capture_wait_s=settle_timeout,
    )

    browser_opts = BrowserCompatOptions(
        allow_autoplay=allow_autoplay,
        hide_overlays=hide_overlays,
        reduced_motion=reduced_motion,
        disable_animations=disable_animations,
        mute_media=mute_media,
        block_media=block_media,
        compatibility_level=forgiveness_level,
        user_agent=override_user_agent,
    )

    runner_opts = RunnerOptions(
        extra_styles=tuple(extra_styles),
        extra_init_scripts=tuple(extra_init_scripts),
        playwright_executable_path=playwright_executable_path,
        extra=extra,
    )

    return ScreenshotOptions(capture=capture_opts, browser=browser_opts, runner=runner_opts)



from __future__ import annotations

from test_utils import create_capture_result

from screenshot import ScreenshotCaptureResult
from screenshot._internal import from_capture_result, per_viewport_resources
from screenshot._shared.errors import ScreenshotError


def test_from_capture_result_builds_resource() -> None:
    capture = create_capture_result(requested=1, captured=1, failed=0)
    result = from_capture_result(
        capture,
        resource_key="key",
        config={"viewports": ["desktop"], "depth": 1, "timeout_s": 30},
    )

    assert result.outcome.name == "SUCCESS"
    assert result.metadata["requested"] == 1


def test_per_viewport_resources_breaks_out_entries() -> None:
    capture = ScreenshotCaptureResult(
        requested=2,
        captured=1,
        failed=1,
        metadata_path=None,
        entries=[
            {"viewport": "desktop", "status": "success"},
            {"viewport": "mobile", "status": "failed", "error": "timeout"},
        ],
        errors=[ScreenshotError(error_type="TimeoutError", message="timeout", retryable=True)],
    )
    resource = from_capture_result(
        capture,
        resource_key="key",
        config={"viewports": ["desktop", "mobile"]},
    )
    viewports = per_viewport_resources(resource)
    assert {v.viewport for v in viewports} == {"desktop", "mobile"}



import argparse
from pathlib import Path

import pytest

from screenshot import cli as screenshot_cli, cli_shared
from screenshot.cli_config import apply_config_to_parser, load_cli_config


def _option_strings(parser: argparse.ArgumentParser) -> set[tuple[str, ...]]:
    opts = set()
    for action in parser._actions:
        if action.option_strings:
            opts.add(tuple(sorted(action.option_strings)))
    return opts


def test_config_matches_programmatic_args(tmp_path: Path) -> None:
    programmatic = argparse.ArgumentParser()
    cli_shared.add_job_arguments(programmatic, include_output_dir=True)

    config_parser = argparse.ArgumentParser()
    cli_shared.add_job_arguments(config_parser, include_output_dir=True, use_config=True)

    assert _option_strings(programmatic) == _option_strings(config_parser)


def test_load_cli_config_custom_path(tmp_path: Path) -> None:
    custom_yaml = tmp_path / "cli.yaml"
    custom_yaml.write_text(
        """
        {
          "version": "0.0.1",
          "metadata": {"package": "screenshot"},
          "arguments": {
            "urls": {
              "flags": ["--urls"],
              "type": "str"
            }
          }
        }
        """,
        encoding="utf-8",
    )
    config = load_cli_config(custom_yaml)
    assert config.version == "0.0.1"
    assert "urls" in config.arguments


def test_apply_config_preset_excludes_advanced() -> None:
    config = load_cli_config()
    parser = argparse.ArgumentParser()
    apply_config_to_parser(parser, config, preset="azure_service")

    options = {opt for action in parser._actions for opt in action.option_strings}
    assert "--extra-css" not in options
    assert "--url" in options or "--urls" in options


def test_missing_config_raises(tmp_path: Path) -> None:
    with pytest.raises(FileNotFoundError):
        load_cli_config(tmp_path / "missing.yaml")


def test_screenshot_cli_respects_config_flag(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setenv("SCREENSHOT_CLI_USE_CONFIG", "1")
    parser = screenshot_cli.build_parser()
    args = parser.parse_args(
        [
            "local",
            "--urls",
            "https://example.com",
            "--output-dir",
            str(tmp_path / "out"),
        ]
    )
    assert args.urls == ["https://example.com"]



import argparse
from pathlib import Path

import pytest

from screenshot import cli as screenshot_cli
from screenshot.cli_shared import add_job_arguments, build_cli_args


def test_add_job_arguments_and_build_cli_args(tmp_path: Path) -> None:
    parser = argparse.ArgumentParser()
    add_job_arguments(parser, include_output_dir=True)

    args = parser.parse_args(
        [
            "--urls",
            "https://example.com",
            "--job-id",
            "home",
            "--partition-date",
            "2025-01-01",
            "--max-pages",
            "4",
            "--depth",
            "2",
            "--viewports",
            "desktop",
            "mobile",
            "--post-nav-wait-s",
            "0.5",
            "--timeout-s",
            "45",
            "--max-retries",
            "5",
            "--job-budget-s",
            "120",
            "--max-viewport-concurrency",
            "2",
            "--scroll",
            "false",
            "--scroll-step-delay-ms",
            "180",
            "--max-scroll-steps",
            "7",
            "--allow-autoplay",
            "false",
            "--hide-overlays",
            "false",
            "--reduced-motion",
            "true",
            "--full-page",
            "false",
            "--mute-media",
            "false",
            "--disable-animations",
            "true",
            "--block-media",
            "true",
            "--pre-capture-wait-s",
            "1.2",
            "--extra-css",
            str(tmp_path / "extra.css"),
            "--extra-js",
            str(tmp_path / "extra.js"),
            "--override-custom-user-agent",
            "MyAgent/5.0",
            "--chromium-compat",
            "medium",
            "--playwright-executable-path",
            "/usr/bin/google-chrome",
            "--output-dir",
            str(tmp_path),
        ]
    )

    (tmp_path / "extra.css").write_text("body{}", encoding="utf-8")
    (tmp_path / "extra.js").write_text("console.log('hi');", encoding="utf-8")

    cli_args = build_cli_args(args)
    assert cli_args.urls == ("https://example.com",)
    assert cli_args.site_ids == ("home",)
    assert cli_args.partition_date == "2025-01-01"
    assert cli_args.post_nav_wait_s == 0.5
    assert cli_args.max_retries == 5
    assert cli_args.job_budget_s == 120.0
    assert cli_args.scroll is False
    assert cli_args.scroll_step_delay_ms == 180
    assert cli_args.max_scroll_steps == 7
    assert cli_args.allow_autoplay is False
    assert cli_args.hide_overlays is False
    assert cli_args.reduced_motion is True
    assert cli_args.full_page is False
    assert cli_args.pre_capture_wait_s == 1.2
    assert cli_args.mute_media is False
    assert cli_args.disable_animations is True
    assert cli_args.block_media is True
    assert cli_args.extra_css_paths
    assert cli_args.extra_js_paths
    assert cli_args.chromium_compat == "medium"
    assert cli_args.max_viewport_concurrency == 2
    assert cli_args.override_custom_user_agent == "MyAgent/5.0"
    assert cli_args.playwright_executable_path == "/usr/bin/google-chrome"


def test_add_job_arguments_without_output_dir(tmp_path: Path) -> None:
    parser = argparse.ArgumentParser()
    add_job_arguments(parser, include_output_dir=False)

    args = parser.parse_args(
        [
            "--urls",
            "https://example.com",
            "--job-id",
            "home",
        ]
    )

    cli_args = build_cli_args(args)

    assert cli_args.urls == ("https://example.com",)
    assert cli_args.site_ids == ("home",)
    assert cli_args.allow_autoplay is None
    assert cli_args.full_page is None
    assert cli_args.chromium_compat is None
    assert cli_args.max_viewport_concurrency is None
    assert cli_args.scroll_step_delay_ms is None
    assert cli_args.max_scroll_steps is None
    assert cli_args.pre_capture_wait_s == 2.5
    assert cli_args.override_custom_user_agent is None
    assert cli_args.playwright_executable_path is None


def test_local_cli_env_flag_disables_config(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setenv("SCREENSHOT_CLI_USE_CONFIG", "0")
    parser = screenshot_cli.build_parser()
    args = parser.parse_args(
        [
            "local",
            "--urls",
            "https://example.com",
            "--output-dir",
            str(tmp_path / "out"),
        ]
    )
    assert isinstance(args, argparse.Namespace)



"""Miscellaneous helpers for CLI processing."""

from __future__ import annotations

import re
from urllib.parse import urlparse

_SAFE_JOB_ID_RE = re.compile(r"[^a-z0-9-]+")


def derive_job_id(url: str, *, fallback: str = "job") -> str:
    """Consistently derive a job identifier from a URL."""

    if url:
        parsed = urlparse(url)
        parts = [parsed.netloc.replace(".", "-").strip("-")]
        if parsed.path and parsed.path not in {"/", ""}:
            segments = [segment for segment in parsed.path.split("/") if segment]
            parts.extend(segments[:2])
        job_id = "-".join(filter(None, parts)).lower()
        job_id = _SAFE_JOB_ID_RE.sub("-", job_id).strip("-")
        if job_id:
            return job_id
    return fallback


__all__ = ["derive_job_id"]



"""Dataclasses describing CLI inputs and intermediate records."""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path

__all__ = ["ScreenshotCliArgs", "RawJobRecord"]


@dataclass(frozen=True)
class ScreenshotCliArgs:
    """Lightweight view of CLI arguments consumed by job collection."""

    input: Path | None
    urls: Sequence[str]
    site_ids: Sequence[str]
    partition_date: str | None
    max_pages: int
    depth: int
    viewports: Sequence[str]
    post_nav_wait_s: float
    timeout_s: float
    max_retries: int | None
    job_budget_s: float | None
    scroll: bool | None = None
    scroll_step_delay_ms: int | None = None
    max_scroll_steps: int | None = None
    allow_autoplay: bool | None = None
    hide_overlays: bool | None = None
    reduced_motion: bool | None = None
    full_page: bool | None = None
    pre_capture_wait_s: float | None = None
    mute_media: bool | None = None
    disable_animations: bool | None = None
    block_media: bool | None = None
    extra_css_paths: Sequence[Path] = ()
    extra_js_paths: Sequence[Path] = ()
    chromium_compat: str | None = None
    max_viewport_concurrency: int | None = None
    override_custom_user_agent: str | None = None
    playwright_executable_path: str | None = None
    backend: str = "playwright"


@dataclass(frozen=True)
class RawJobRecord:
    """Raw job description before options are materialised."""

    job_id: str
    url: str
    backend: str
    partition_date: str | None
    html_snapshot_path: str | None
    metadata: dict[str, object]
    options_override: dict[str, object] | None



from pathlib import Path
from typing import NoReturn

import pytest

from screenshot._models_plans import BrowserPlan, CapturePlan, RunnerPlan
from screenshot.models import ScreenshotOptions
from screenshot.selenium_runner import (
    _apply_browser_mutations,
    _capture_full_page,
    _capture_url_with_viewports,
    _collect_links,
    _failed_entry,
    _scroll_page,
    _SeleniumImportError,
    capture_screenshots_async,
)


@pytest.mark.asyncio
async def test_selenium_runner_missing_dependency(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    def _raise() -> NoReturn:
        raise _SeleniumImportError("Selenium backend unavailable")

    monkeypatch.setattr("screenshot.selenium_runner._load_selenium_modules", _raise)

    result = await capture_screenshots_async(
        "test-job",
        "https://example.com",
        store_dir=tmp_path,
        partition_date=None,
        options=ScreenshotOptions(),
        html_snapshot_path=None,
        cancel_token=None,
        storage=None,
    )

    assert result.requested == 0
    assert result.captured == 0
    assert result.failed == 0
    assert result.errors and "Selenium backend unavailable" in result.errors[0].message


@pytest.mark.asyncio
async def test_selenium_navigation_retry_logs(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture
) -> None:
    caplog.set_level("DEBUG")

    # Build fakes returned by _load_selenium_modules()
    class TimeoutException(Exception):
        pass

    class WebDriverException(Exception):
        pass

    class Options:
        def __init__(self) -> None:
            self.binary_location = None

        def add_argument(self, *_args: object) -> None:
            return None

    class _Driver:
        def __init__(self) -> None:
            self._calls = 0
            self.title = "ok"

        def set_page_load_timeout(self, _timeout: object) -> None:
            return None

        def get(self, _url: object) -> None:
            # Fail first call with timeout, then succeed
            self._calls += 1
            if self._calls == 1:
                raise TimeoutException("nav timeout")
            return None

        def execute_cdp_cmd(self, *_args: object, **_kwargs: object) -> dict[str, str]:
            # Return values used by full-page capture path if invoked
            return {"data": ""}

        def get_screenshot_as_png(self) -> bytes:
            return b"PNG"

        @property
        def page_source(self) -> str:
            return "<html/>"

        def quit(self) -> None:
            return None

    class webdriver:
        @staticmethod
        def Chrome(*_args: object, **_kwargs: object) -> _Driver:
            return _Driver()

    def _fake_loader() -> (
        tuple[type[webdriver], type[Options], type[TimeoutException], type[WebDriverException]]
    ):
        return webdriver, Options, TimeoutException, WebDriverException

    monkeypatch.setattr("screenshot.selenium_runner._load_selenium_modules", _fake_loader)

    res = await capture_screenshots_async(
        "job",
        "https://example.com",
        store_dir=tmp_path,
        partition_date=None,
        options=ScreenshotOptions(),
        html_snapshot_path=None,
        cancel_token=None,
        storage=None,
    )

    assert res.captured >= 1
    # Ensure our retry logging hook fired
    assert "Retrying navigation (selenium)" in caplog.text


def test_apply_browser_mutations_executes_optional_scripts() -> None:
    executed: list[str] = []

    class Driver:
        def execute_script(self, script: str, *args: object) -> None:
            executed.append(script)

    browser_plan = BrowserPlan(
        allow_autoplay=True,
        hide_overlays=True,
        reduced_motion=True,
        disable_animations=True,
        mute_media=True,
        block_media=False,
        compatibility_level="low",
        user_agent="UA",
    )
    runner_plan = RunnerPlan(
        extra_styles=("body {color:red;}",),
        extra_init_scripts=("console.log('init');",),
        navigation_strategies=(),
        playwright_executable_path=None,
        extra={},
    )

    _apply_browser_mutations(Driver(), browser_plan, runner_plan)

    assert any("autoplay" in script.lower() for script in executed)
    assert any("hide-overlays" in script for script in executed)
    assert any("disable-animations" in script for script in executed)
    assert any("console.log('init');" in script for script in executed)


def test_scroll_page_and_collect_links(monkeypatch: pytest.MonkeyPatch) -> None:
    heights = [100, 120, 120]

    class Driver:
        def __init__(self) -> None:
            self._scroll_calls = 0

        def execute_script(self, script: str, *args: object) -> int | bool | None:
            if "Math.max" in script:
                value = heights[min(self._scroll_calls, len(heights) - 1)]
                self._scroll_calls += 1
                return value
            if "window.scrollY" in script:
                return True
            return None

    monkeypatch.setattr("time.sleep", lambda _seconds: None)
    driver = Driver()
    scroll_height = _scroll_page(driver, max_steps=3, delay_ms=10)
    assert scroll_height == heights[-1]

    class LinkDriver:
        def execute_script(self, *_args: object) -> list[str | None]:
            return ["https://example.com/a", "/b", "mailto:test@example.com", None]

    links = _collect_links(
        driver=LinkDriver(),
        current_url="https://example.com",
        origin="example.com",
        depth=0,
        max_depth=1,
        max_pages=5,
    )
    assert links == ["https://example.com/a", "https://example.com/b"]


def test_failed_entry_structure() -> None:
    entry = _failed_entry(2, 1, "desktop", "https://example.com", 10.0, "boom")
    assert entry["index"] == 2
    assert entry["status"] == "failed"
    assert entry["timeout_seconds"] == 10.0


def test_capture_url_handles_driver_init_failure(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    capture_plan = CapturePlan(
        max_pages=1,
        depth=0,
        delay_ms=0,
        viewport_specs={"desktop": {"viewport": {"width": 100, "height": 50}}},
        viewport_details=[
            {
                "name": "desktop",
                "width": 100,
                "height": 50,
                "device_scale_factor": 1.0,
                "is_mobile": False,
            }
        ],
        primary_viewport="desktop",
        timeout_sec=5.0,
        max_total_duration_sec=5.0,
        max_capture_attempts=1,
        max_viewport_concurrency=1,
        scroll_enabled=False,
        scroll_step_delay_ms=250,
        max_scroll_steps=15,
        full_page_capture=False,
        settle_timeout_ms=0,
    )
    browser_plan = BrowserPlan(
        allow_autoplay=False,
        hide_overlays=False,
        reduced_motion=False,
        disable_animations=False,
        mute_media=False,
        block_media=False,
        compatibility_level="low",
        user_agent=None,
    )
    runner_plan = RunnerPlan(
        extra_styles=(),
        extra_init_scripts=(),
        navigation_strategies=(),
        playwright_executable_path=None,
        extra={},
    )

    def fake_create_driver(**_kwargs: object) -> NoReturn:
        raise RuntimeError("boom")

    monkeypatch.setattr("screenshot.selenium_runner._create_driver", fake_create_driver)

    class DummyWebdriver:
        pass

    class DummyOptions:
        pass

    outcome = _capture_url_with_viewports(
        webdriver=DummyWebdriver,
        Options=DummyOptions,
        TimeoutException=RuntimeError,
        WebDriverException=RuntimeError,
        job_id="job",
        url="https://example.com",
        origin="example.com",
        depth=0,
        capture_plan=capture_plan,
        browser_plan=browser_plan,
        runner_plan=runner_plan,
        screenshot_dir=tmp_path,
        page_index=0,
        snapshot_path=None,
        snapshot_written=False,
    )

    assert outcome.failed == 1
    assert outcome.errors and outcome.errors[0].error_type == "driver_init"


def test_capture_url_uses_scroll_tuning(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    recorded: dict[str, int] = {}

    def fake_scroll_page(_driver: object, *, max_steps: int, delay_ms: int) -> int:
        recorded["max_steps"] = max_steps
        recorded["delay_ms"] = delay_ms
        return 200

    monkeypatch.setattr("screenshot.selenium_runner._scroll_page", fake_scroll_page)

    class FakeDriver:
        title = "Fake"

        def set_page_load_timeout(self, *_args: object, **_kwargs: object) -> None:
            return None

        def get(self, *_args: object, **_kwargs: object) -> None:
            return None

        def execute_script(self, *_args: object, **_kwargs: object) -> int:
            return 0

        def get_screenshot_as_png(self) -> bytes:
            return b"PNG"

        def quit(self) -> None:
            return None

    def fake_create_driver(**_kwargs: object) -> tuple[FakeDriver, None]:
        return FakeDriver(), None

    monkeypatch.setattr("screenshot.selenium_runner._create_driver", fake_create_driver)

    capture_plan = CapturePlan(
        max_pages=1,
        depth=0,
        delay_ms=0,
        viewport_specs={"desktop": {"viewport": {"width": 100, "height": 50}}},
        viewport_details=[
            {
                "name": "desktop",
                "width": 100,
                "height": 50,
                "device_scale_factor": 1.0,
                "is_mobile": False,
            }
        ],
        primary_viewport="desktop",
        timeout_sec=5.0,
        max_total_duration_sec=5.0,
        max_capture_attempts=1,
        max_viewport_concurrency=1,
        scroll_enabled=True,
        scroll_step_delay_ms=155,
        max_scroll_steps=6,
        full_page_capture=False,
        settle_timeout_ms=0,
    )
    browser_plan = BrowserPlan(
        allow_autoplay=False,
        hide_overlays=False,
        reduced_motion=False,
        disable_animations=False,
        mute_media=False,
        block_media=False,
        compatibility_level="low",
        user_agent=None,
    )
    runner_plan = RunnerPlan(
        extra_styles=(),
        extra_init_scripts=(),
        navigation_strategies=(),
        playwright_executable_path=None,
        extra={},
    )

    class DummyWebdriver:
        pass

    class DummyOptions:
        pass

    outcome = _capture_url_with_viewports(
        webdriver=DummyWebdriver,
        Options=DummyOptions,
        TimeoutException=RuntimeError,
        WebDriverException=RuntimeError,
        job_id="job",
        url="https://example.com",
        origin="example.com",
        depth=0,
        capture_plan=capture_plan,
        browser_plan=browser_plan,
        runner_plan=runner_plan,
        screenshot_dir=tmp_path,
        page_index=0,
        snapshot_path=None,
        snapshot_written=False,
    )

    assert outcome.captured == 1
    assert recorded == {"max_steps": 6, "delay_ms": 155}


def test_selenium_timings_emitted_when_enabled(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setenv("SCREENSHOT_ENABLE_TIMING", "1")
    monkeypatch.setattr("screenshot.selenium_runner.ENABLE_TIMING", True)

    class FakeDriver:
        title = "Fake"

        def set_page_load_timeout(self, *_args: object, **_kwargs: object) -> None:
            return None

        def get(self, *_args: object, **_kwargs: object) -> None:
            return None

        def execute_script(self, *_args: object, **_kwargs: object) -> None:
            return None

        def get_screenshot_as_png(self) -> bytes:
            return b"PNG"

        def quit(self) -> None:
            return None

    def fake_create_driver(**_kwargs: object) -> tuple[FakeDriver, None]:
        return FakeDriver(), None

    monkeypatch.setattr("screenshot.selenium_runner._create_driver", fake_create_driver)

    capture_plan = CapturePlan(
        max_pages=1,
        depth=0,
        delay_ms=0,
        viewport_specs={"desktop": {"viewport": {"width": 100, "height": 50}}},
        viewport_details=[
            {
                "name": "desktop",
                "width": 100,
                "height": 50,
                "device_scale_factor": 1.0,
                "is_mobile": False,
            }
        ],
        primary_viewport="desktop",
        timeout_sec=5.0,
        max_total_duration_sec=5.0,
        max_capture_attempts=1,
        max_viewport_concurrency=1,
        scroll_enabled=False,
        scroll_step_delay_ms=250,
        max_scroll_steps=15,
        full_page_capture=False,
        settle_timeout_ms=0,
    )
    browser_plan = BrowserPlan(
        allow_autoplay=False,
        hide_overlays=False,
        reduced_motion=False,
        disable_animations=False,
        mute_media=False,
        block_media=False,
        compatibility_level="low",
        user_agent=None,
    )
    runner_plan = RunnerPlan(
        extra_styles=(),
        extra_init_scripts=(),
        navigation_strategies=(),
        playwright_executable_path=None,
        extra={},
    )

    class DummyWebdriver:
        pass

    class DummyOptions:
        pass

    outcome = _capture_url_with_viewports(
        webdriver=DummyWebdriver,
        Options=DummyOptions,
        TimeoutException=RuntimeError,
        WebDriverException=RuntimeError,
        job_id="job",
        url="https://example.com",
        origin="example.com",
        depth=0,
        capture_plan=capture_plan,
        browser_plan=browser_plan,
        runner_plan=runner_plan,
        screenshot_dir=tmp_path,
        page_index=0,
        snapshot_path=None,
        snapshot_written=False,
    )

    assert outcome.captured == 1
    assert outcome.entries
    timings = outcome.entries[0].get("timings")
    assert isinstance(timings, dict)
    assert {
        "navigation_ms",
        "scroll_ms",
        "post_scroll_wait_ms",
        "screenshot_ms",
        "total_ms",
    }.issubset(timings.keys())
    assert timings["scroll_ms"] == 0.0
    assert timings["post_scroll_wait_ms"] == 0.0


def test_capture_full_page_fallback(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    class Driver:
        def __init__(self) -> None:
            self._scripts: list[str] = []

        def execute_cdp_cmd(self, *_args: object, **_kwargs: object) -> None:
            raise RuntimeError("cdp unsupported")

        def get_screenshot_as_png(self) -> bytes:
            return b"FAKEPNG"

        def execute_script(self, *_args: object, **_kwargs: object) -> int:
            return 200

    out_path = tmp_path / "shot.png"
    _capture_full_page(
        Driver(),
        {"viewport": {"width": 120, "height": 80}, "device_scale_factor": 1.0},
        out_path,
    )
    assert out_path.exists()



"""Derived plan dataclasses consumed by capture backends along with serialisation helpers."""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass

from ._models_options import OPTIONS_SCHEMA_VERSION


@dataclass(frozen=True, kw_only=True, slots=True)
class CapturePlan:
    """Sanitized capture plan consumed by screenshot runners."""

    max_pages: int
    depth: int
    delay_ms: int
    viewport_specs: dict[str, dict[str, object]]
    viewport_details: list[dict[str, object]]
    primary_viewport: str
    timeout_sec: float
    max_total_duration_sec: float
    max_capture_attempts: int
    max_viewport_concurrency: int
    scroll_enabled: bool
    scroll_step_delay_ms: int
    max_scroll_steps: int
    full_page_capture: bool
    settle_timeout_ms: int


@dataclass(frozen=True, kw_only=True, slots=True)
class BrowserPlan:
    """Browser compatibility directives produced by options sanitiser."""

    allow_autoplay: bool
    hide_overlays: bool
    reduced_motion: bool
    disable_animations: bool
    mute_media: bool
    block_media: bool
    compatibility_level: str
    user_agent: str | None


@dataclass(frozen=True, kw_only=True, slots=True)
class RunnerPlan:
    """Runtime directives (styles/scripts/extra payload) for runner launch."""

    extra_styles: Sequence[str]
    extra_init_scripts: Sequence[str]
    navigation_strategies: Sequence[dict[str, object]]
    playwright_executable_path: str | None
    extra: dict[str, object]


@dataclass(frozen=True, kw_only=True, slots=True)
class SanitizedPlans:
    """Container bundling capture/browser/runner plans for runners."""

    capture: CapturePlan
    browser: BrowserPlan
    runner: RunnerPlan

    def to_dict(self) -> dict[str, object]:
        return {
            "schema_version": OPTIONS_SCHEMA_VERSION,
            "capture": {
                "max_pages": self.capture.max_pages,
                "depth": self.capture.depth,
                "delay_ms": self.capture.delay_ms,
                "viewport_specs": dict(self.capture.viewport_specs),
                "viewport_details": list(self.capture.viewport_details),
                "primary_viewport": self.capture.primary_viewport,
                "timeout_sec": self.capture.timeout_sec,
                "max_total_duration_sec": self.capture.max_total_duration_sec,
                "max_capture_attempts": self.capture.max_capture_attempts,
                "max_viewport_concurrency": self.capture.max_viewport_concurrency,
                "scroll_enabled": self.capture.scroll_enabled,
                "scroll_step_delay_ms": self.capture.scroll_step_delay_ms,
                "max_scroll_steps": self.capture.max_scroll_steps,
                "full_page_capture": self.capture.full_page_capture,
                "settle_timeout_ms": self.capture.settle_timeout_ms,
            },
            "browser": {
                "allow_autoplay": self.browser.allow_autoplay,
                "hide_overlays": self.browser.hide_overlays,
                "reduced_motion": self.browser.reduced_motion,
                "disable_animations": self.browser.disable_animations,
                "mute_media": self.browser.mute_media,
                "block_media": self.browser.block_media,
                "compatibility_level": self.browser.compatibility_level,
                "user_agent": self.browser.user_agent,
            },
            "runner": {
                "extra_styles": list(self.runner.extra_styles),
                "extra_init_scripts": list(self.runner.extra_init_scripts),
                "navigation_strategies": list(self.runner.navigation_strategies),
                "playwright_executable_path": self.runner.playwright_executable_path,
                "extra": dict(self.runner.extra),
            },
        }



from __future__ import annotations

from pathlib import Path

import pytest

from screenshot._models_options import (
    OPTIONS_SCHEMA_VERSION,
    CaptureOptions,
    ScreenshotOptions,
    _capture_from_nested_dict,
    _capture_to_dict,
    _runner_from_nested_dict,
    _to_int,
)


def test_capture_options_validation_and_merge() -> None:
    defaults = {"desktop": {"viewport": {"width": 800, "height": 600}}}
    options = CaptureOptions(enabled=True, viewports=("desktop", {"name": "custom"}))
    merged = options.merged_viewports(defaults)

    assert merged["desktop"]["viewport"]["width"] == 800
    assert "custom" in merged

    with pytest.raises(ValueError):
        CaptureOptions(max_pages=0)
    with pytest.raises(TypeError):
        CaptureOptions(viewports=("desktop", 123))  # type: ignore[arg-type]


def test_capture_from_dict_and_round_trip() -> None:
    capture = _capture_from_nested_dict(
        {
            "enabled": True,
            "max_pages": "2",
            "depth": "1",
            "viewports": ["mobile"],
            "post_nav_wait_s": "1.25",
            "timeout_s": "30",
            "max_total_duration_s": "0",
            "max_capture_attempts": "4",
            "max_viewport_concurrency": "3",
            "scroll": True,
            "scroll_step_delay_ms": "120",
            "max_scroll_steps": "5",
            "full_page": True,
            "pre_capture_wait_s": "0.75",
        }
    )
    as_dict = _capture_to_dict(capture)
    assert as_dict["max_pages"] == 2
    assert as_dict["depth"] == 1
    assert as_dict["viewports"] == ["mobile"]
    assert as_dict["max_total_duration_s"] is None
    assert as_dict["pre_capture_wait_s"] == 0.75
    assert as_dict["scroll_step_delay_ms"] == 120
    assert as_dict["max_scroll_steps"] == 5


def test_runner_from_nested_dict_coerces_types() -> None:
    runner = _runner_from_nested_dict(
        {
            "extra_styles": ["a"],
            "extra_init_scripts": ("b",),
            "playwright_executable_path": Path("/tmp/chrome"),
            "extra": {"feature": "on"},
        }
    )

    assert runner.playwright_executable_path == str(Path("/tmp/chrome"))
    assert runner.extra_styles == ("a",)
    assert runner.extra["feature"] == "on"


def test_screenshot_options_schema_and_serialization() -> None:
    opts = ScreenshotOptions.from_dict({"capture": {"enabled": True}})
    payload = opts.to_dict()
    assert payload["schema_version"] == OPTIONS_SCHEMA_VERSION
    assert payload["capture"]["enabled"] is True

    with pytest.raises(ValueError):
        ScreenshotOptions.from_dict({})


def test_to_int_accepts_bool() -> None:
    assert _to_int(True, default=0) == 1



"""Async-first orchestration helpers for screenshot batches.

The `ScreenshotService` wraps a backend implementation and coordinates
concurrency, cancellation, and aggregation of `ScreenshotCaptureResult`
objects. It is primarily used by CLIs and worker processes that want a
consistent interface regardless of the underlying runner.

Example:
    >>> import asyncio
    >>> from pathlib import Path
    >>> from screenshot.cli import PlaywrightScreenshotBackend
    >>> service = ScreenshotService(backend=PlaywrightScreenshotBackend(), concurrency=2)
    >>> jobs = [ScreenshotJob(job_id="demo", url="https://example.com", output_root=Path("/tmp"))]
    >>> batch = asyncio.run(service.capture_async(jobs))
    >>> batch.successes[0].captured
    1
"""

from __future__ import annotations

import asyncio
from collections.abc import Iterable
from typing import Generic, Protocol, TypeVar, cast, runtime_checkable

from ._shared.errors import make_error
from .models import ScreenshotBatchResult, ScreenshotCaptureResult, ScreenshotJob

TResult = TypeVar("TResult", bound=ScreenshotCaptureResult)
TResult_co = TypeVar("TResult_co", bound=ScreenshotCaptureResult, covariant=True)


@runtime_checkable
class CancelToken(Protocol):
    """Protocol describing the cancel token hook used by runners."""

    def raise_if_cancelled(self) -> None: ...


@runtime_checkable
class ScreenshotBackend(Protocol, Generic[TResult_co]):
    """Backend interface responsible for executing individual screenshot jobs."""

    async def run_job_async(self, job: ScreenshotJob) -> TResult_co: ...


class ScreenshotService(Generic[TResult]):
    """High-level helper that executes batches of screenshot jobs."""

    def __init__(self, backend: ScreenshotBackend[TResult], *, concurrency: int = 1) -> None:
        if concurrency <= 0:
            raise ValueError("concurrency must be positive")
        self._backend = backend
        self._default_concurrency = concurrency

    async def capture_async(
        self,
        jobs: Iterable[ScreenshotJob],
        *,
        concurrency: int | None = None,
        cancel_token: CancelToken | None = None,
    ) -> ScreenshotBatchResult:
        """Execute the provided jobs asynchronously with optional overrides.

        Args:
            jobs: Iterable of screenshot jobs to process.
            concurrency: Optional override for the maximum in-flight jobs.
            cancel_token: Cooperative cancellation token exposing `raise_if_cancelled`.

        Returns:
            Aggregated `ScreenshotBatchResult` summarizing successes/failures.

        Example:
            >>> import asyncio
            >>> from pathlib import Path
            >>> async def _demo():
            ...     backend = PlaywrightScreenshotBackend()  # doctest: +SKIP
            ...     service = ScreenshotService(backend, concurrency=2)  # doctest: +SKIP
            ...     jobs = [
            ...         ScreenshotJob(
            ...             job_id="demo",
            ...             url="https://example.com",
            ...             output_root=Path("/tmp"),
            ...         )
            ...     ]  # doctest: +SKIP
            ...     return await service.capture_async(jobs)  # doctest: +SKIP
            >>> # asyncio.run(_demo())  # doctest: +SKIP
        """

        job_list: list[ScreenshotJob] = list(jobs)
        if not job_list:
            return ScreenshotBatchResult(results=[])

        limit = concurrency or self._default_concurrency
        semaphore = asyncio.Semaphore(limit)
        results: list[TResult] = []

        async def _run(job: ScreenshotJob) -> None:
            if cancel_token is not None:
                cancel_token.raise_if_cancelled()
            async with semaphore:
                try:
                    result = await self._backend.run_job_async(job)
                except asyncio.CancelledError:
                    raise
                except Exception as exc:  # pragma: no cover - defensive
                    error_result = ScreenshotCaptureResult(
                        requested=1,
                        captured=0,
                        failed=1,
                        metadata_path=None,
                        entries=[],
                        errors=[
                            make_error(
                                "backend",
                                f"Screenshot job failed: {exc}",
                                retryable=True,
                                details={"job_id": job.job_id},
                            )
                        ],
                        job_id=job.job_id,
                    )
                    results.append(cast(TResult, error_result))
                else:
                    if result.job_id is None:
                        result.job_id = job.job_id
                    results.append(result)

        await asyncio.gather(*(_run(job) for job in job_list))
        return ScreenshotBatchResult.from_results(results)

    def capture(
        self,
        jobs: Iterable[ScreenshotJob],
        *,
        concurrency: int | None = None,
    ) -> ScreenshotBatchResult:
        """Synchronous wrapper around `capture_async` for CLI scripts.

        This helper spins up a dedicated event loop via `asyncio.run`, so it
        should only be used from non-async contexts such as entry points or
        small scripts. Call `capture_async` when integrating into async code.

        Example:
            >>> service = ScreenshotService(backend=...)  # doctest: +SKIP
            >>> service.capture([])  # doctest: +SKIP
            ScreenshotBatchResult(results=[])
        """

        return asyncio.run(self.capture_async(jobs, concurrency=concurrency))



"""Public API models for screenshot capture.

This module exposes the core public types needed by consumers of the screenshot
package. Internal implementation details are kept in private _models_* modules.
"""

from __future__ import annotations

# Public configuration types
# Internal types needed by some tests/advanced users
# These are intentionally not in __all__ to discourage casual use
from ._models_options import (
    CaptureOptions,
    ScreenshotOptions,
)

# Public result types
from ._models_results import (
    ScreenshotBatchResult,
    ScreenshotCaptureResult,
    ScreenshotJob,
    ScreenshotResourceError,
    ScreenshotResourceOutcome,
    ScreenshotResourceResult,
)

__all__ = [
    # Core configuration
    "ScreenshotOptions",
    "CaptureOptions",
    # Job specification
    "ScreenshotJob",
    # Result types
    "ScreenshotCaptureResult",
    "ScreenshotBatchResult",
    "ScreenshotResourceResult",
    "ScreenshotResourceOutcome",
    "ScreenshotResourceError",
]



"""Structured error types shared within the screenshot package."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum


@dataclass(frozen=True)
class ScreenshotError:
    """Typed error emitted during screenshot capture."""

    error_type: str
    message: str
    retryable: bool = False
    details: dict[str, object] = field(default_factory=dict)

    def to_dict(self) -> dict[str, object]:
        return {
            "error_type": self.error_type,
            "message": self.message,
            "retryable": self.retryable,
            "details": dict(self.details),
        }

    def __str__(self) -> str:
        return self.message


def make_error(
    error_type: str,
    message: str,
    *,
    retryable: bool = False,
    details: dict[str, object] | None = None,
) -> ScreenshotError:
    """Factory that normalises inputs into a ScreenshotError."""

    return ScreenshotError(
        error_type=error_type,
        message=message,
        retryable=retryable,
        details=dict(details or {}),
    )


class ErrorCategory(str, Enum):
    TIMEOUT = "screenshot.timeout"
    NAVIGATION = "screenshot.navigation"
    BROWSER = "screenshot.browser"
    STORAGE = "screenshot.storage"
    UNKNOWN = "screenshot.unknown"

    @classmethod
    def from_error_type(cls, error_type: str | None) -> ErrorCategory:
        normalized = (error_type or "").lower().strip()
        if not normalized:
            return cls.UNKNOWN
        if "timeout" in normalized:
            return cls.TIMEOUT
        if "navigation" in normalized:
            return cls.NAVIGATION
        if "browser" in normalized:
            return cls.BROWSER
        if "storage" in normalized:
            return cls.STORAGE
        return cls.UNKNOWN



"""Storage abstractions shared within the screenshot package."""

from __future__ import annotations

import asyncio
from collections.abc import Callable
from enum import Enum
from functools import lru_cache
from pathlib import Path
from typing import Any, Protocol, runtime_checkable


@lru_cache(maxsize=1)
def _ensure_parent_impl() -> Callable[[Path], None]:
    from infra_core.fs_utils import ensure_parent as infra_ensure_parent

    return infra_ensure_parent


def _ensure_parent(path: Path) -> None:
    _ensure_parent_impl()(path)


def _default_json_serializer(value: object) -> Any:
    if isinstance(value, Path):
        return str(value)
    if hasattr(value, "isoformat"):
        try:
            return value.isoformat()
        except TypeError:
            pass
    if isinstance(value, Enum):
        return value.value
    return value


@runtime_checkable
class StorageBackend(Protocol):
    """Writable storage interface for screenshot artifacts."""

    def ensure_parent(self, path: Path) -> None: ...

    def write_json(self, path: Path, data: dict[str, Any]) -> Path: ...

    def write_text(self, path: Path, text: str) -> Path: ...


@runtime_checkable
class CloudStorageBackend(StorageBackend, Protocol):
    """Extension for backends that support synchronous uploads."""

    def upload_file(self, path: Path) -> None: ...


@runtime_checkable
class AsyncStorageBackend(StorageBackend, Protocol):
    """Extension for backends that expose true async uploads."""

    async def upload_file_async(self, path: Path) -> None: ...


class LocalStorageBackend:
    """Default storage backend that only writes to the local filesystem."""

    def ensure_parent(self, path: Path) -> None:
        _ensure_parent(path)

    def write_json(self, path: Path, data: dict[str, Any]) -> Path:
        self.ensure_parent(path)
        path.write_text(dumps_json(data), encoding="utf-8")
        return path

    async def write_json_async(self, path: Path, data: dict[str, Any]) -> Path:
        return await asyncio.to_thread(self.write_json, path, data)

    def write_text(self, path: Path, text: str) -> Path:
        self.ensure_parent(path)
        path.write_text(text, encoding="utf-8")
        return path

    async def write_text_async(self, path: Path, text: str) -> Path:
        return await asyncio.to_thread(self.write_text, path, text)

    def upload_file(self, path: Path) -> None:  # pragma: no cover - no-op default
        return None


def dumps_json(data: dict[str, Any]) -> str:
    """Serialize data to JSON with stable formatting and Path/Enum awareness."""

    import json

    text = json.dumps(
        data,
        ensure_ascii=False,
        indent=2,
        sort_keys=True,
        default=_default_json_serializer,
    )
    return f"{text}\n"



from __future__ import annotations

import json
import os
import threading
import time
from collections.abc import Generator
from http.server import BaseHTTPRequestHandler, HTTPServer
from pathlib import Path
from typing import Any
from urllib.parse import urlparse

import pytest

from screenshot import (
    ScreenshotBackend,
    ScreenshotBatchResult,
    ScreenshotCaptureResult,
    ScreenshotJob,
    ScreenshotOptions,
    ScreenshotService,
)
from screenshot.models import CaptureOptions
from screenshot.playwright_runner import capture_screenshots_async as playwright_capture_async

SITES_UNDER_TEST = [
    "https://landonorris.com/",
    "https://wonjyou.studio/",
    "https://algon.iq/",
]

SYSTEM_CHROME_PATH = "/usr/bin/google-chrome-stable"

_LOCAL_SITE_PAGES = {
    "/": (
        "<html><head><title>Local Home</title></head>"
        "<body><h1>It works!</h1><p id='content'>Local content</p></body></html>"
    ),
    "/gallery": (
        "<html><body><section><h2>Gallery</h2><img src='one.png'/>"
        "<img src='two.png'/></section></body></html>"
    ),
    "/slow": "<html><body><p>Slow page</p></body></html>",
}


def _require_e2e() -> None:
    if os.getenv("RUN_E2E") != "1":
        pytest.skip("Set RUN_E2E=1 to enable screenshot service E2E tests")


def _require_playwright() -> None:
    try:
        import playwright  # noqa: F401
    except ImportError as exc:  # pragma: no cover - optional dependency
        pytest.skip(f"Playwright is not installed ({exc}); skipping screenshot E2E test")


def _require_real_sites() -> None:
    if os.getenv("RUN_REAL_SITES") != "1":
        pytest.skip("Set RUN_REAL_SITES=1 to enable external site captures")


@pytest.fixture(autouse=True)
def _use_system_chrome(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("PLAYWRIGHT_EXECUTABLE_PATH", SYSTEM_CHROME_PATH)


def _job_id_from_url(url: str) -> str:
    parsed = urlparse(url)
    parts = [parsed.netloc.replace(".", "-").strip("-")]
    if parsed.path and parsed.path not in {"/", ""}:
        segments = [segment for segment in parsed.path.split("/") if segment]
        parts.extend(segments[:2])
    job_id = "-".join(filter(None, parts)) or "site"
    return job_id.lower()


class _PlaywrightBackend(ScreenshotBackend[ScreenshotCaptureResult]):
    """Lightweight backend that forwards jobs to the Playwright runner."""

    async def run_job_async(self, job: ScreenshotJob) -> ScreenshotCaptureResult:
        result = await playwright_capture_async(
            job.job_id,
            job.url,
            store_dir=job.output_root,
            partition_date=job.partition_date,
            options=job.options,
            html_snapshot_path=job.html_snapshot_path,
            cancel_token=job.cancel_token,
        )
        if result.job_id is None:
            result.job_id = job.job_id
        return result


def _make_capture_options(**overrides: Any) -> CaptureOptions:
    """Create CaptureOptions with defaults and overrides.

    Args:
        **overrides: Keyword arguments matching CaptureOptions fields

    Returns:
        CaptureOptions instance with merged settings
    """
    defaults: dict[str, Any] = {
        "enabled": True,
        "max_pages": 1,
        "depth": 0,
        "viewports": ("desktop",),
        "post_nav_wait_s": 0.2,
        "timeout_s": 15.0,
        "max_capture_attempts": 1,
    }
    # Merge overrides with defaults
    defaults.update(overrides)

    # Create CaptureOptions, letting it validate the types
    return CaptureOptions(**defaults)


def _make_screenshot_options(**capture_overrides: Any) -> ScreenshotOptions:
    """Create ScreenshotOptions with default capture settings.

    Args:
        **capture_overrides: Keyword arguments for CaptureOptions

    Returns:
        ScreenshotOptions instance configured for testing
    """
    options = ScreenshotOptions(capture=_make_capture_options(**capture_overrides))
    options.runner.playwright_executable_path = SYSTEM_CHROME_PATH
    return options


def _assert_successful_capture(result: ScreenshotCaptureResult) -> None:
    assert result.succeeded, f"Screenshot capture failed for {result.job_id}: {result.errors}"
    assert result.captured >= 1, f"No screenshots captured for {result.job_id}"
    assert result.metadata_path is not None, f"Missing metadata path for {result.job_id}"
    metadata_path = result.metadata_path
    assert metadata_path.exists(), f"Metadata file missing at {metadata_path}"

    payload = json.loads(metadata_path.read_text(encoding="utf-8"))
    entries = payload.get("entries") if isinstance(payload, dict) else None
    assert entries, f"No metadata entries recorded for {result.job_id}"

    site_root = metadata_path.parent.parent
    screenshot_paths = []
    for entry in entries:
        if not isinstance(entry, dict):
            continue
        rel_path = entry.get("screenshot_path")
        status = entry.get("status")
        if rel_path and status in {"captured", "success"}:
            candidate = Path(rel_path)
            if not candidate.is_absolute():
                candidate = site_root / rel_path
            screenshot_paths.append(candidate)

    assert screenshot_paths, f"No captured screenshot paths found for {result.job_id}"
    for screenshot_path in screenshot_paths:
        assert screenshot_path.exists(), f"Screenshot file missing at {screenshot_path}"


class _LocalSiteHandler(BaseHTTPRequestHandler):
    def do_GET(self) -> None:  # noqa: N802 - BaseHTTPRequestHandler signature
        if self.path == "/slow":
            time.sleep(1.5)
        body = _LOCAL_SITE_PAGES.get(self.path)
        if body is None:
            self.send_response(404)
            self.send_header("Content-Type", "text/plain; charset=utf-8")
            self.end_headers()
            self.wfile.write(b"not found")
            return
        self.send_response(200)
        self.send_header("Content-Type", "text/html; charset=utf-8")
        self.end_headers()
        self.wfile.write(body.encode("utf-8"))

    def log_message(self, format: str, *args: object) -> None:  # noqa: A003
        return  # Silence per-test HTTP logging noise


@pytest.fixture(scope="module")
def local_test_site() -> Generator[str, None, None]:
    server = HTTPServer(("127.0.0.1", 0), _LocalSiteHandler)
    thread = threading.Thread(target=server.serve_forever, daemon=True)
    thread.start()
    base_url = f"http://127.0.0.1:{server.server_address[1]}"
    try:
        yield base_url
    finally:
        server.shutdown()
        thread.join()


@pytest.mark.asyncio
async def test_screenshot_service_captures_local_site(tmp_path: Path, local_test_site: str) -> None:
    _require_e2e()
    _require_playwright()

    backend = _PlaywrightBackend()
    service = ScreenshotService(backend, concurrency=1)

    job = ScreenshotJob(
        job_id="local-home",
        url=f"{local_test_site}/",
        output_root=tmp_path,
        options=_make_screenshot_options(),
    )

    batch_result = await service.capture_async([job])
    assert len(batch_result.results) == 1
    _assert_successful_capture(batch_result.results[0])


@pytest.mark.asyncio
async def test_screenshot_service_captures_multiple_jobs(
    tmp_path: Path, local_test_site: str
) -> None:
    _require_e2e()
    _require_playwright()

    backend = _PlaywrightBackend()
    service = ScreenshotService(backend, concurrency=2)

    jobs = [
        ScreenshotJob(
            job_id="local-home",
            url=f"{local_test_site}/",
            output_root=tmp_path,
            options=_make_screenshot_options(),
        ),
        ScreenshotJob(
            job_id="local-gallery",
            url=f"{local_test_site}/gallery",
            output_root=tmp_path,
            options=_make_screenshot_options(post_nav_wait_s=0.3),
        ),
    ]

    batch_result = await service.capture_async(jobs)
    assert len(batch_result.results) == 2
    job_ids = {result.job_id for result in batch_result.results}
    assert job_ids == {"local-home", "local-gallery"}
    for result in batch_result.results:
        _assert_successful_capture(result)


@pytest.mark.asyncio
async def test_screenshot_service_scroll_tuning(tmp_path: Path, local_test_site: str) -> None:
    _require_e2e()
    _require_playwright()

    backend = _PlaywrightBackend()
    service = ScreenshotService(backend, concurrency=1)

    job = ScreenshotJob(
        job_id="local-scroll-tuning",
        url=f"{local_test_site}/",
        output_root=tmp_path,
        options=_make_screenshot_options(
            scroll=True,
            scroll_step_delay_ms=0,
            max_scroll_steps=1,
            pre_capture_wait_s=0.0,
        ),
    )

    batch_result = await service.capture_async([job])
    assert len(batch_result.results) == 1
    _assert_successful_capture(batch_result.results[0])


@pytest.mark.asyncio
async def test_screenshot_service_reports_failures(tmp_path: Path, local_test_site: str) -> None:
    _require_e2e()
    _require_playwright()

    backend = _PlaywrightBackend()
    service = ScreenshotService(backend, concurrency=2)

    good_job = ScreenshotJob(
        job_id="local-home",
        url=f"{local_test_site}/",
        output_root=tmp_path,
        options=_make_screenshot_options(),
    )
    failing_job = ScreenshotJob(
        job_id="bad-host",
        url="http://127.0.0.1:9/nope",
        output_root=tmp_path,
        options=_make_screenshot_options(timeout_s=5.0, max_capture_attempts=1),
    )

    batch_result = await service.capture_async([good_job, failing_job])
    assert len(batch_result.results) == 2
    successes = [result for result in batch_result.results if result.job_id == "local-home"]
    failures = [result for result in batch_result.results if result.job_id == "bad-host"]
    assert len(successes) == 1
    assert len(failures) == 1
    _assert_successful_capture(successes[0])
    failure = failures[0]
    assert not failure.succeeded
    assert failure.metadata_path and failure.metadata_path.exists()
    entry_errors = [
        entry.get("error")
        for entry in failure.entries
        if isinstance(entry, dict) and entry.get("error")
    ]
    assert failure.errors or entry_errors, "Expected failure to record error details"


@pytest.mark.asyncio
async def test_screenshot_service_captures_real_sites(tmp_path: Path) -> None:
    """End-to-end validation that the screenshot service captures real sites."""

    _require_e2e()
    _require_playwright()
    _require_real_sites()

    backend = _PlaywrightBackend()
    service = ScreenshotService(backend, concurrency=1)

    jobs: list[ScreenshotJob] = []
    for url in SITES_UNDER_TEST:
        job_id = _job_id_from_url(url)
        options = _make_screenshot_options(
            enabled=True,
            max_pages=1,
            depth=0,
            viewports=("desktop",),
            post_nav_wait_s=2.0,
            timeout_s=90.0,
            max_capture_attempts=2,
        )
        jobs.append(
            ScreenshotJob(
                job_id=job_id,
                url=url,
                output_root=tmp_path,
                options=options,
            )
        )

    batch_result: ScreenshotBatchResult = await service.capture_async(jobs)
    assert len(batch_result.results) == len(jobs)

    for result in batch_result.results:
        _assert_successful_capture(result)



"""Configuration-driven argument definitions for screenshot CLIs."""

from __future__ import annotations

import argparse
import json
from collections.abc import Callable, Iterable
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Final

_DEFAULT_CONFIG_PATH = Path(__file__).resolve().parent / "config" / "cli_v1.json"


class _MissingType:
    """Sentinel that marks argument defaults as unspecified."""


_Missing: Final = _MissingType()


@dataclass(frozen=True)
class CLIArgumentSpec:
    """Specification for a single CLI argument."""

    arg_id: str
    dest: str
    flags: tuple[str, ...]
    type: str
    default: Any = field(default_factory=lambda: _Missing)
    help: str = ""
    action: str | None = None
    choices: tuple[str, ...] | None = None
    required: bool = False
    metavar: str | None = None
    nargs: str | int | None = None
    const: Any | None = None
    parser_type: str = "standard"
    deprecated: bool = False
    added_in: str | None = None


@dataclass(frozen=True)
class CLIArgumentGroup:
    """Logical grouping of related arguments."""

    name: str
    description: str
    arguments: tuple[str, ...]


@dataclass(frozen=True)
class CLIServicePreset:
    """Pre-configured argument selections for common entrypoints."""

    name: str
    description: str
    include_groups: tuple[str, ...] = ()
    exclude_args: tuple[str, ...] = ()
    additional_args: tuple[str, ...] = ()


@dataclass(frozen=True)
class CLIConfigSchema:
    """Top-level configuration container."""

    version: str
    package: str
    arguments: dict[str, CLIArgumentSpec]
    argument_groups: dict[str, CLIArgumentGroup] = field(default_factory=dict)
    service_presets: dict[str, CLIServicePreset] = field(default_factory=dict)


def load_cli_config(config_path: Path | str | None = None) -> CLIConfigSchema:
    """Load CLI configuration from YAML or embedded defaults."""

    if config_path is None:
        target = _DEFAULT_CONFIG_PATH
    else:
        target = Path(config_path)
    if not target.exists():
        raise FileNotFoundError(f"CLI config not found: {target}")
    with target.open("r", encoding="utf-8") as handle:
        data = json.load(handle)
    return _parse_config(data)


def apply_config_to_parser(
    parser: argparse.ArgumentParser,
    config: CLIConfigSchema,
    *,
    include_groups: Iterable[str] | None = None,
    exclude_args: Iterable[str] | None = None,
    preset: str | None = None,
) -> None:
    """Apply the configuration to an ``argparse.ArgumentParser`` instance."""

    from ._shared.cli.arguments import _parse_bool  # noqa: PLC0415 - avoid circular import

    if preset:
        if preset not in config.service_presets:
            raise ValueError(
                f"Unknown preset '{preset}'. Available presets: {sorted(config.service_presets)}"
            )
        preset_cfg = config.service_presets[preset]
        include_groups = preset_cfg.include_groups
        exclude_args = (*preset_cfg.exclude_args, *(exclude_args or ()))
        additional_args = preset_cfg.additional_args
    else:
        additional_args = ()

    included_args: set[str] = set(additional_args)
    if include_groups is None:
        included_args.update(config.arguments.keys())
    else:
        for group_name in include_groups:
            group = config.argument_groups.get(group_name)
            if group is None:
                raise ValueError(f"Unknown argument group '{group_name}'")
            included_args.update(group.arguments)

    if exclude_args:
        included_args -= set(exclude_args)

    for arg_id in included_args:
        if arg_id not in config.arguments:
            raise ValueError(f"Argument '{arg_id}' not defined in CLI config")
        spec = config.arguments[arg_id]
        kwargs: dict[str, Any] = {}
        if spec.help:
            kwargs["help"] = spec.help
        if spec.choices:
            kwargs["choices"] = list(spec.choices)
        if spec.required:
            kwargs["required"] = True
        if spec.metavar:
            kwargs["metavar"] = spec.metavar
        if spec.nargs is not None:
            kwargs["nargs"] = spec.nargs
        if spec.const is not None:
            kwargs["const"] = spec.const
        if spec.action:
            kwargs["action"] = spec.action
        if spec.default is not _Missing:
            kwargs["default"] = spec.default

        if spec.parser_type == "custom_bool":
            kwargs["type"] = _parse_bool
        else:
            type_callable = _resolve_type(spec.type)
            if type_callable is not None and spec.action not in {"store_true", "store_false"}:
                kwargs["type"] = type_callable

        parser.add_argument(*spec.flags, dest=spec.dest, **kwargs)


TypeConverter = Callable[[str], object]


def _resolve_type(type_name: str | None) -> TypeConverter | None:
    if type_name is None:
        return None
    converters: dict[str, TypeConverter] = {
        "int": int,
        "float": float,
        "str": str,
        "path": Path,
        "bool": bool,
    }
    return converters.get(type_name)


def _parse_config(data: dict[str, Any]) -> CLIConfigSchema:
    version = data.get("version")
    if not version:
        raise ValueError("Config missing 'version' field")
    metadata = data.get("metadata") or {}
    package = metadata.get("package", "unknown")

    arguments: dict[str, CLIArgumentSpec] = {}
    for arg_id, arg_data in (data.get("arguments") or {}).items():
        flags = tuple(arg_data["flags"])
        dest = arg_data.get("dest", _derive_dest(flags[0]))
        default = arg_data.get("default", _Missing)
        if default is None and "default" not in arg_data:
            default = _Missing
        arguments[arg_id] = CLIArgumentSpec(
            arg_id=arg_id,
            dest=dest,
            flags=flags,
            type=arg_data.get("type", "str"),
            default=default,
            help=arg_data.get("help", ""),
            action=arg_data.get("action"),
            choices=tuple(arg_data["choices"]) if arg_data.get("choices") else None,
            required=arg_data.get("required", False),
            metavar=arg_data.get("metavar"),
            nargs=arg_data.get("nargs"),
            const=arg_data.get("const"),
            parser_type=arg_data.get("parser_type", "standard"),
            deprecated=arg_data.get("deprecated", False),
            added_in=arg_data.get("added_in"),
        )

    argument_groups = {
        name: CLIArgumentGroup(
            name=name,
            description=group_data.get("description", ""),
            arguments=tuple(group_data.get("arguments", ())),
        )
        for name, group_data in (data.get("argument_groups") or {}).items()
    }

    service_presets = {
        name: CLIServicePreset(
            name=name,
            description=preset.get("description", ""),
            include_groups=tuple(preset.get("include_groups", ())),
            exclude_args=tuple(preset.get("exclude_args", ())),
            additional_args=tuple(preset.get("additional_args", ())),
        )
        for name, preset in (data.get("service_presets") or {}).items()
    }

    return CLIConfigSchema(
        version=version,
        package=package,
        arguments=arguments,
        argument_groups=argument_groups,
        service_presets=service_presets,
    )


def _derive_dest(flag: str) -> str:
    return flag.lstrip("-").replace("-", "_")



"""Parsing helpers that fan out CLI arguments into raw job records."""

from __future__ import annotations

import json
from collections.abc import Sequence
from pathlib import Path

from .schema import RawJobRecord, ScreenshotCliArgs
from .utils import derive_job_id

__all__ = ["load_cli_inputs"]


def _load_snippets(paths: Sequence[Path]) -> list[str]:
    snippets: list[str] = []
    for path in paths:
        snippets.append(path.read_text(encoding="utf-8"))
    return snippets


def load_cli_inputs(args: ScreenshotCliArgs) -> tuple[list[RawJobRecord], list[str], list[str]]:
    """Return raw job records plus pre-loaded CSS/JS snippets."""

    backend = (args.backend or "playwright").strip().lower()
    css_snippets = _load_snippets(args.extra_css_paths)
    js_snippets = _load_snippets(args.extra_js_paths)

    records: list[RawJobRecord] = []

    if args.input:
        with args.input.open("r", encoding="utf-8") as handle:
            for line in handle:
                if not line.strip():
                    continue
                record = json.loads(line)
                url = (record.get("url") or "").strip()
                if not url:
                    continue
                job_id = (
                    record.get("job_id") or record.get("site_id") or ""
                ).strip() or derive_job_id(url)
                partition = record.get("partition_date") or args.partition_date
                overrides = (
                    record.get("options") if isinstance(record.get("options"), dict) else None
                )
                snapshot = record.get("html_snapshot_path")
                snapshot_path = str(snapshot).strip() if snapshot else None
                metadata = record.get("metadata")
                metadata_dict = dict(metadata) if isinstance(metadata, dict) else {}
                metadata_dict.setdefault("source", "cli")
                records.append(
                    RawJobRecord(
                        job_id=job_id,
                        url=url,
                        backend=backend,
                        partition_date=partition,
                        html_snapshot_path=snapshot_path,
                        metadata=metadata_dict,
                        options_override=overrides,
                    )
                )

    urls = [url.strip() for url in (args.urls or []) if url and url.strip()]
    site_ids = [sid.strip() for sid in (args.site_ids or []) if sid and sid.strip()]
    if urls:
        for index, url in enumerate(urls):
            fallback = f"site-{index + 1}"
            job_id = (
                site_ids[index] if index < len(site_ids) else derive_job_id(url, fallback=fallback)
            )
            records.append(
                RawJobRecord(
                    job_id=job_id,
                    url=url,
                    backend=backend,
                    partition_date=args.partition_date,
                    html_snapshot_path=None,
                    metadata={"source": "cli", "input": "args"},
                    options_override=None,
                )
            )

    return records, css_snippets, js_snippets
